"""IPMI module.

Todo:
    replace with pyghmi.

"""
import logging
from datetime import timedelta
from subprocess import PIPE, CalledProcessError, run
from typing import Dict, List, Tuple

from spicerack.decorators import retry
from spicerack.exceptions import SpicerackCheckError, SpicerackError

IPMI_PASSWORD_MAX_LEN: int = 20
IPMI_PASSWORD_MIN_LEN: int = 16
IPMI_SAFE_BOOT_PARAMS: Tuple[str, ...] = (
    "0000000000",  # No overrides
    "8000000000",  # Boot Flag Valid
    "8000020000",  # Boot Flag Valid and Screen blank
)
logger = logging.getLogger(__name__)


class IpmiError(SpicerackError):
    """Custom exception class for errors of the Ipmi class."""


class IpmiCheckError(SpicerackCheckError):
    """Custom exception class for check errors of the Ipmi class."""


class Ipmi:
    """Class to manage remote IPMI via ipmitool."""

    def __init__(self, mgmt_fqdn: str, password: str, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            mgmt_fqdn (str): the management console FQDN to target.
            password (str): the password to use to connect via IPMI.
            dry_run (bool, optional): whether this is a DRY-RUN.

        """
        self.env: Dict[str, str] = {"IPMITOOL_PASSWORD": password}
        self._mgmt_fqdn = mgmt_fqdn
        self._dry_run = dry_run

    def command(self, command_parts: List[str], is_safe: bool = False, hide_parts: Tuple = ()) -> str:
        """Run an ipmitool command for a remote management console FQDN.

        Arguments:
            command_parts (list): a list of :py:class:`str` with the IPMI command components to execute.
            is_safe (bool, optional): if this is a safe command to run also in DRY RUN mode.
            hide_parts (tuple, optional): tuple with indexes of the command_parts list that should be redacted in logs
                and outputs because contain sensitive data. For example setting it to (2, 4) would replace in logs and
                outputs the 3rd and 5th element of the command_parts list.

        Returns:
            str: the output of the ipmitool command.

        Raises:
            spicerack.ipmi.IpmiError: on failure.

        """
        command = [
            "ipmitool",
            "-I",
            "lanplus",
            "-H",
            self._mgmt_fqdn,
            "-U",
            "root",
            "-E",
        ]
        redacted_parts = command_parts[:]
        for i in hide_parts:
            redacted_parts[i] = "__REDACTED__"

        logger.info("Running IPMI command: %s", " ".join(command + redacted_parts))

        if self._dry_run and not is_safe:
            return ""

        try:
            output = run(command + command_parts, env=self.env.copy(), stdout=PIPE, check=True).stdout.decode()
        except CalledProcessError as e:
            raise IpmiError(f"Remote IPMI for {self._mgmt_fqdn} failed (exit={e.returncode}): {e.output}") from e

        logger.debug(output)

        return output

    def check_connection(self) -> None:
        """Ensure that remote IPMI is working for the management console FQDN.

        Raises:
            spicerack.ipmi.IpmiError: if unable to connect or execute a test command.

        """
        self.power_status()

    def power_status(self) -> str:
        """Get the current power status for the management console FQDN.

        Raises:
            spicerack.ipmi.IpmiError: if unable to get the power status.

        """
        identifier = "Chassis Power is "
        status = self.command(["chassis", "power", "status"], is_safe=True)
        if not status.startswith(identifier):
            raise IpmiError(f"Unexpected chassis status: {status}")

        return status[len(identifier) :].strip()

    def check_bootparams(self) -> None:
        """Check if the BIOS boot parameters are back to normal values.

        Raises:
            spicerack.ipmi.IpmiCheckError: if the BIOS boot parameters are incorrect.

        """
        param = self._get_boot_parameter("Boot parameter data")
        if param not in IPMI_SAFE_BOOT_PARAMS:
            raise IpmiCheckError(f"Expected BIOS boot params in {IPMI_SAFE_BOOT_PARAMS} got: {param}")

    @retry(
        tries=3,
        delay=timedelta(seconds=20),
        backoff_mode="linear",
        exceptions=(IpmiCheckError,),
    )
    def force_pxe(self) -> None:
        """Force PXE for the next boot and verify that the setting was applied.

        Raises:
            spicerack.ipmi.IpmiCheckError: if unable to verify the PXE mode within the retries.

        """
        self.command(["chassis", "bootparam", "set", "bootflag", "force_pxe", "options=reset"])
        boot_device = self._get_boot_parameter("Boot Device Selector")
        if boot_device != "Force PXE":
            message = "Unable to verify that Force PXE is set. The host might reboot in the current OS"
            if self._dry_run:
                logger.warning(message)
            else:
                raise IpmiCheckError(message)

    def remove_boot_override(self) -> None:
        """Remove any boot override, if present for the next boot and verify that the change was applied.

        Raises:
            spicerack.ipmi.IpmiCheckError: if unable to verify the boot mode.

        """
        self.command(["chassis", "bootparam", "set", "bootflag", "none", "options=reset"])
        boot_device = self._get_boot_parameter("Boot Device Selector")
        if boot_device != "No override":
            message = "Unable to verify that the boot override was removed. The host might reboot in PXE"
            if self._dry_run:
                logger.warning(message)
            else:
                raise IpmiCheckError(message)

    def reboot(self) -> None:
        """Reboot a host via IPMI, either performing a power cycle or a power on based on the power status."""
        status = self.power_status()
        if status == "off":
            operation = "on"
        else:
            operation = "cycle"

        self.command(["chassis", "power", operation])

    def _get_boot_parameter(self, param_label: str) -> str:
        """Get a specific boot parameter of the host.

        Arguments:
            param_label (str): the label of the boot parameter to lookout for.

        Raises:
            spicerack.ipmi.IpmiError: if unable to find the given label or to extract its value.

        Returns:
            str: the value of the parameter.

        """
        bootparams = self.command(["chassis", "bootparam", "get", "5"], is_safe=True)
        for line in bootparams.splitlines():
            if param_label in line:
                try:
                    value = line.split(":")[1].strip(" \n")
                    break
                except IndexError as e:
                    raise IpmiError(f"Unable to extract value for parameter '{param_label}' from line: {line}") from e
        else:
            raise IpmiError(f"Unable to find the boot parameter '{param_label}' in: {bootparams}")
        return value

    @staticmethod
    def _get_password_store_size(password: str) -> int:
        """Parse the password to determine the correct storage size.

        Ipmitool stores passwords in either 16 or 20 byte strings depending
        on the password length.

        Arguments:
            password(str): the password string to parse

        Raises:
            spicerack.ipmi.IpmiError: if unable password is too big or too small

        Returns:
            int: A number representing the storage size

        """
        if len(password) > IPMI_PASSWORD_MAX_LEN:
            raise IpmiError(f"New passwords is greater then the {IPMI_PASSWORD_MAX_LEN} byte limit")
        if len(password) > IPMI_PASSWORD_MIN_LEN:
            return IPMI_PASSWORD_MAX_LEN
        if len(password) == IPMI_PASSWORD_MIN_LEN:
            return IPMI_PASSWORD_MIN_LEN
        raise IpmiError(f"New passwords must be {IPMI_PASSWORD_MIN_LEN} bytes minimum")

    def reset_password(self, username: str, password: str) -> None:
        """Reset the given usernames password to the one provided.

        Arguments:
            username (str): The username who's password will be reset must not be empty
            password (str): The new password, length between :py:const:`spicerack.ipmi.IPMI_PASSWORD_MIN_LEN` and
             :py:const:`spicerack.ipmi.IPMI_PASSWORD_MAX_LEN` bytes

        Raises:
            spicerack.ipmi.IpmiError: if unable reset password or arguments invalid

        """
        if not username:
            raise IpmiError("Username can not be an empty string")

        password_store_size = str(Ipmi._get_password_store_size(password))

        try:
            user_id = self._get_user_id(username)
        except IpmiError:
            # some systems (HP?) use channel 2
            logger.info("unable to find user in channel 1 testing channel 2")
            user_id = self._get_user_id(username, 2)

        success = f"Set User Password command successful (user {user_id})\n"
        result = self.command(["user", "set", "password", user_id, password, password_store_size], hide_parts=(4,))

        if self._dry_run:
            return

        if result != success:
            raise IpmiError(f"Password reset failed for username: {username}")

        if username == "root":
            current_password = self.env["IPMITOOL_PASSWORD"]
            self.env["IPMITOOL_PASSWORD"] = password
            try:
                self.check_connection()
            except IpmiError as error:
                self.env["IPMITOOL_PASSWORD"] = current_password
                raise IpmiError("Password reset failed for username: root") from error

    def _get_user_id(self, username: str, channel: int = 1) -> str:
        """Get the user ID associated with a given username.

        Arguments:
            username (str): The username to search for
            channel (int): The channel number for the user list; Default: 1

        Raises:
            spicerack.ipmi.IpmiError: if unable to find the given username.

        Returns:
            str: the user ID associated with the username

        """
        userlist = self.command(["user", "list", str(channel)], is_safe=True)
        for line in userlist.splitlines():
            words = line.split()
            if words[0] == "ID":
                continue
            if words[1] == username:
                return words[0]
        raise IpmiError(f"Unable to find ID for username: {username}")



"""Alertmanager module tests."""
import json
from unittest import mock

import pytest
import requests

from spicerack import peeringdb
from spicerack.tests import get_fixture_path


def test_pdb_proxies():
    """Test thet the proxies are set correctly."""
    proxies = {"http": "http://proxy:8080"}
    pdb = peeringdb.PeeringDB(proxies=proxies)
    assert pdb.session.proxies == proxies


class TestPeeringdb:
    """Tests for the Peeringdb class."""

    @pytest.fixture(autouse=True)
    def setup_method(self, tmp_path):
        """Setup test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.base_url = peeringdb.PeeringDB.baseurl
        self.pdb = peeringdb.PeeringDB(token="foo")
        self.pdb_cached = peeringdb.PeeringDB(token="foo", cachedir=tmp_path / "cache")
        self.pdb_cached_low_ttl = peeringdb.PeeringDB(token="foo", cachedir=tmp_path / "cache", ttl=1)
        # load test fixtures
        self.asn = json.loads(get_fixture_path("peeringdb", "asn.json").read_text())
        self.ixlan = json.loads(get_fixture_path("peeringdb", "ixlan.json").read_text())

    def test_pdb_fetch_net(self, requests_mock):
        """Test fetching net resource."""
        requests_mock.get(f"{self.base_url}net/1", json=self.asn)
        assert self.pdb_cached.fetch("net", 1) == self.asn["data"]
        # Trigger the caching
        assert self.pdb_cached.fetch("net", 1) == self.asn["data"]
        # Ensure we only made one request
        assert len(requests_mock.request_history) == 1

    def test_pdb_fetch_asn_cache(self, requests_mock):
        """PeringDB.fetch() should return a dict."""
        requests_mock.get(f"{self.base_url}net", json=self.asn)
        assert self.pdb_cached.fetch_asn(14907) == self.asn["data"]
        # Trigger the caching
        assert self.pdb_cached.fetch_asn(14907) == self.asn["data"]
        assert len(requests_mock.request_history) == 1

    def test_pdb_fetch_asn_no_cache(self, requests_mock):
        """PeringDB.fetch() should return a dict."""
        requests_mock.get(f"{self.base_url}net", json=self.asn)
        assert self.pdb.fetch_asn(14907) == self.asn["data"]
        assert self.pdb.fetch_asn(14907) == self.asn["data"]
        assert len(requests_mock.request_history) == 2

    def test_pdb_fetch_no_resource_id_no_cache(self, requests_mock):
        """PeringDB.fetch() should return a dict."""
        requests_mock.get(self.base_url + "ixlan", json=self.ixlan)
        assert self.pdb.fetch("ixlan") == self.ixlan["data"]
        assert self.pdb.fetch("ixlan") == self.ixlan["data"]
        assert len(requests_mock.request_history) == 2

    def test_pdb_fetch_no_resource_id_cache(self, requests_mock):
        """PeringDB.fetch() should return a dict."""
        # Trigger the caching
        requests_mock.get(self.base_url + "ixlan", json=self.ixlan)
        assert self.pdb_cached.fetch("ixlan") == self.ixlan["data"]
        assert self.pdb_cached.fetch("ixlan") == self.ixlan["data"]
        assert len(requests_mock.request_history) == 1

    @mock.patch("spicerack.peeringdb.time.time")
    def test_pdb_fetch_cache_age(self, mock_time, requests_mock):
        """PeringDB.fetch() should return a dict."""
        mock_time.return_value = 2147483647  # epoch end date ensures the cache is expired
        requests_mock.get(self.base_url + "ixlan", json=self.ixlan)
        assert self.pdb_cached_low_ttl.fetch("ixlan") == self.ixlan["data"]
        assert self.pdb_cached_low_ttl.fetch("ixlan") == self.ixlan["data"]
        assert len(requests_mock.request_history) == 2

    def test_pdb_fetch_not_found(self, requests_mock):
        """PeringDB.fetch() should return a dict."""
        pdb = peeringdb.PeeringDB()
        requests_mock.get(self.base_url + "ixlan", text="", status_code=requests.codes["not_found"])

        with pytest.raises(peeringdb.PeeringDBError, match=r"Server response with status.*"):
            pdb.fetch("ixlan")
        assert len(requests_mock.request_history) == 1

    def test_pdb_cache_key(self):
        """Peeringdb get_cache_key should return the correct key."""
        # pylint: disable=protected-access
        # useful to test this function even though its private
        assert self.pdb._get_cache_key("net") == "net/index"
        assert self.pdb._get_cache_key("net", resource_id=1) == "net/1"
        assert self.pdb._get_cache_key("net", filters={"asn": 42, "depth": 2}) == "net/asn/42/depth/2"
        assert self.pdb._get_cache_key("net", resource_id=1, filters={"asn": 42, "depth": 2}) == "net/1/asn/42/depth/2"



"""Netbox module tests."""
import logging
from copy import deepcopy
from datetime import datetime
from unittest import mock

import pytest
import requests

from spicerack import redfish

ACCOUNTS_RESPONSE = {
    "@odata.context": "/redfish/v1/$metadata#ManagerAccountCollection.ManagerAccountCollection",
    "@odata.id": "/redfish/v1/AccountService/Accounts",
    "@odata.type": "#ManagerAccountCollection.ManagerAccountCollection",
    "Description": "BMC User Accounts Collection",
    "Members": [
        {"@odata.id": "/redfish/v1/AccountService/Accounts/1"},
        {"@odata.id": "/redfish/v1/AccountService/Accounts/2"},
        {"@odata.id": "/redfish/v1/AccountService/Accounts/3"},
    ],
    "Members@odata.count": 3,
    "Name": "Accounts Collection",
}
ACCOUNT_RESPONSE = {
    "@odata.context": "/redfish/v1/$metadata#ManagerAccount.ManagerAccount",
    "@odata.id": "/redfish/v1/AccountService/Accounts/{user_id}",
    "@odata.type": "#ManagerAccount.v1_5_0.ManagerAccount",
    "AccountTypes": ["Redfish", "SNMP", "OEM"],
    "Description": "User Account",
    "Enabled": True,
    "Id": "{user_id}",
    "Links": {"Role": {"@odata.id": "/redfish/v1/AccountService/Roles/Administrator"}},
    "Locked": False,
    "Name": "User Account",
    "OEMAccountTypes": ["IPMI", "SOL", "WSMAN", "UI", "RACADM"],
    "Password": None,
    "PasswordChangeRequired": False,
    "RoleId": "Administrator",
    "SNMP": {
        "AuthenticationKey": None,
        "AuthenticationKeySet": True,
        "AuthenticationProtocol": "HMAC_SHA96",
        "EncryptionKey": None,
        "EncryptionKeySet": False,
        "EncryptionProtocol": "CFB128_AES128",
    },
    "UserName": "username",
}
DELL_SCP = {
    "SystemConfiguration": {
        "Comments": [{"Comment": "First comment"}],
        "Components": [
            {
                "Attributes": [
                    {
                        "Comment": "Read and Write",
                        "Name": "Some.Attribute.2",
                        "Set On Import": "False",
                        "Value": "value",
                    },
                    {
                        "Comment": "Read and Write",
                        "Name": "Some.Attribute.1",
                        "Set On Import": "False",
                        "Value": "value",
                    },
                ],
                "FQDD": "Some.Component.2",
            },
            {
                "Attributes": [
                    {
                        "Comment": "Read and Write",
                        "Name": "Some.Attribute.1",
                        "Set On Import": "False",
                        "Value": "value",
                    },
                ],
                "FQDD": "Some.Component.1",
            },
            {
                "Attributes": [
                    {
                        "Comment": "Read and Write",
                        "Name": "Comma.Separated.List.1",
                        "Set On Import": "False",
                        "Value": "value1,value2",
                    },
                    {
                        "Comment": "Read and Write",
                        "Name": "Comma.Space.Separated.List.1",
                        "Set On Import": "False",
                        "Value": "value1, value2",
                    },
                ],
                "FQDD": "List.Component.1",
            },
        ],
        "Model": "PowerEdge R440",
        "ServiceTag": "12ABC34",
        "TimeStamp": "Thu Dec  9 09:32:06 2021",
    },
}
DELL_TASK_REPONSE = {
    "@odata.context": "/redfish/v1/$metadata#Task.Task",
    "@odata.id": "/redfish/v1/TaskService/Tasks/JID_1234567890",
    "@odata.type": "#Task.v1_1_1.Task",
    "Description": "Server Configuration and other Tasks running on iDRAC are listed here",
    "EndTime": "TIME_NA",
    "Id": "JID_1234567890",
    "Messages": [
        {
            "Oem": {
                "Custom": "Structure",
            },
        },
        {
            "Message": "Exporting Server Configuration Profile.",
            "MessageArgs": [],
            "MessageArgs@odata.count": 0,
            "MessageId": "SYS057",
        },
        {
            "Message": "Successfully exported Server Configuration Profile",
            "MessageArgs": [],
            "MessageArgs@odata.count": 0,
            "MessageID": "SYS043",
        },
    ],
    "Messages@odata.count": 1,
    "Name": "Export: Server Configuration Profile",
    "StartTime": "2021-12-09T14:36:29-06:00",
    "TaskState": "Running",
    "TaskStatus": "OK",
}
DELL_TASK_REPONSE_EPOC = deepcopy(DELL_TASK_REPONSE)
DELL_TASK_REPONSE_EPOC["EndTime"] = "1969-12-31T18:00:00-06:00"
DELL_TASK_REPONSE_BAD_TIME = deepcopy(DELL_TASK_REPONSE)
DELL_TASK_REPONSE_BAD_TIME["EndTime"] = "bad value"
MODEL_RESPONSE = {
    "@odata.context": "/redfish/v1/$metadata#Manager.Manager",
    "@odata.id": "/redfish/v1/Managers/iDRAC.Embedded.1",
    "@odata.type": "#Manager.v1_9_0.Manager",
    "Model": "14G Monolithic",
}
MODEL_RESPONSE_BAD = deepcopy(MODEL_RESPONSE)
MODEL_RESPONSE_BAD["Model"] = "Foobar"
PUSHURI_RESPONSE = {
    "@odata.context": "/redfish/v1/$metadata#UpdateService.UpdateService",
    "@odata.id": "/redfish/v1/UpdateService",
    "@odata.type": "#UpdateService.v1_8_1.UpdateService",
    "HttpPushUri": "/redfish/v1/UpdateService/FirmwareInventory",
}
LCLOG_RESPONSE = {
    "@odata.context": "/redfish/v1/$metadata#LogEntryCollection.LogEntryCollection",
    "@odata.id": "/redfish/v1/Managers/iDRAC.Embedded.1/LogServices/Lclog/Entries",
    "@odata.type": "#LogEntryCollection.LogEntryCollection",
    "Description": "LC Logs for this manager",
    "Members": [
        {
            "@odata.id": "/redfish/v1/Managers/iDRAC.Embedded.1/LogServices/Lclog/Entries/1735",
            "@odata.type": "#LogEntry.v1_6_1.LogEntry",
            "Created": "2022-06-22T17:01:17-05:00",
            "Description": "Log Entry 1735",
            "EntryType": "Oem",
            "Id": "1735",
            "Links": {"OriginOfCondition": {"@odata.id": "/redfish/v1/Managers/iDRAC.Embedded.1"}},
            "Message": "Successfully logged in using root, from " "10.192.32.49 and REDFISH.",
            "MessageArgs": ["root", "10.192.32.49", "REDFISH"],
            "MessageArgs@odata.count": 3,
            "MessageId": "USR0030",
            "Name": "Log Entry 1735",
            "Oem": {
                "Dell": {
                    "@odata.type": "#DellLCLogEntry.v1_0_0.DellLCLogEntry",
                    "Category": "Audit",
                    "Comment": None,
                    "LastUpdatedByUser": None,
                }
            },
            "OemRecordFormat": "Dell",
            "Severity": "OK",
        },
        {
            "@odata.id": "/redfish/v1/Managers/iDRAC.Embedded.1/LogServices/Lclog/Entries/1734",
            "@odata.type": "#LogEntry.v1_6_1.LogEntry",
            "Created": "2022-06-22T16:58:17-05:00",
            "Description": "Log Entry 1734",
            "EntryType": "Oem",
            "Id": "1734",
            "Links": {},
            "Message": "The (installation or configuration) job " "JID_559308065328 is successfully completed.",
            "MessageArgs": ["JID_559308065328"],
            "MessageArgs@odata.count": 1,
            "MessageId": "JCP037",
            "Name": "Log Entry 1734",
            "Oem": {
                "Dell": {
                    "@odata.type": "#DellLCLogEntry.v1_0_0.DellLCLogEntry",
                    "Category": "Configuration",
                    "Comment": None,
                    "LastUpdatedByUser": None,
                }
            },
            "OemRecordFormat": "Dell",
            "Severity": "OK",
        },
        {
            "@odata.id": "/redfish/v1/Managers/iDRAC.Embedded.1/LogServices/Lclog/Entries/1692",
            "@odata.type": "#LogEntry.v1_6_1.LogEntry",
            "Created": "2022-06-22T16:42:55-05:00",
            "Description": "Log Entry 1692",
            "EntryType": "Oem",
            "Id": "1692",
            "Links": {},
            "Message": "The iDRAC firmware was rebooted with the " "following reason: user initiated.",
            "MessageArgs": ["user initiated"],
            "MessageArgs@odata.count": 1,
            "MessageId": "RAC0182",
            "Name": "Log Entry 1692",
            "Oem": {
                "Dell": {
                    "@odata.type": "#DellLCLogEntry.v1_0_0.DellLCLogEntry",
                    "Category": "Audit",
                    "Comment": None,
                    "LastUpdatedByUser": None,
                }
            },
            "OemRecordFormat": "Dell",
            "Severity": "OK",
        },
    ],
    "Members@odata.count": 1735,
    "Members@odata.nextLink": "/redfish/v1/Managers/iDRAC.Embedded.1/LogServices/Lclog/Entries?$skip=50",
    "Name": "Log Entry Collection",
}

LCLOG_RESPONSE_NO_MESSAGE = deepcopy(LCLOG_RESPONSE)
LCLOG_RESPONSE_NO_MESSAGE["Members"] = []


def add_accounts_mock_responses(requests_mock):
    """Setup requests mock URLs and return payloads for all the existing users."""
    requests_mock.get("/redfish/v1/Managers/iDRAC.Embedded.1/Accounts", json=ACCOUNTS_RESPONSE)
    users = {"1": "user", "2": "root", "3": "guest"}
    for user_id, username in users.items():
        response = deepcopy(ACCOUNT_RESPONSE)
        response["Id"] = response["Id"].format(user_id=user_id)
        response["@odata.id"] = response["@odata.id"].format(user_id=user_id)
        response["UserName"] = username
        requests_mock.get(
            f"/redfish/v1/AccountService/Accounts/{user_id}", json=response, headers={"ETag": f"12345-{user_id}"}
        )


class TestRedfish:
    """Tests for the Redfish class."""

    @pytest.fixture(autouse=True)
    def setup_method(self, requests_mock):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        self.redfish = redfish.Redfish("test.example.org", "root", "mysecret", dry_run=False)
        self.redfish_dry_run = redfish.Redfish("test.example.org", "root", "mysecret", dry_run=True)
        self.requests_mock = requests_mock

    def test_property_magic_str(self):
        """It should equal the fqdn."""
        assert str(self.redfish) == "root@test.example.org"

    def test_property_fqdn(self):
        """It should equal the fqdn."""
        assert self.redfish.fqdn == "test.example.org"

    @pytest.mark.parametrize("response, generation", ((MODEL_RESPONSE, 14), (MODEL_RESPONSE_BAD, 1)))
    def test_property_generation(self, response, generation):
        """It should return the generation."""
        self.requests_mock.get("/redfish/v1/Managers/iDRAC.Embedded.1?$select=Model", json=response)
        assert self.redfish.generation == generation
        # assert twice to check cached version
        assert self.redfish.generation == generation

    def test_property_pushuri(self):
        """It should return the pushuri."""
        self.requests_mock.get("/redfish/v1/UpdateService?$select=HttpPushUri", json=PUSHURI_RESPONSE)
        assert self.redfish.pushuri == "/redfish/v1/UpdateService/FirmwareInventory"
        # assert twice to check cached version
        assert self.redfish.pushuri == "/redfish/v1/UpdateService/FirmwareInventory"

    def test_most_recent_member(self):
        """It should return the item with the most recent date."""
        data = [
            {"date": "2022-01-01T00:00:00-00:00"},
            {"date": "1970-01-01T00:00:00-00:00"},
            {"date": "1971-01-01T00:00:00-00:00"},
        ]
        assert self.redfish.most_recent_member(data, "date") == {"date": "2022-01-01T00:00:00-00:00"}

    @pytest.mark.parametrize(
        "response, reboot_time",
        ((LCLOG_RESPONSE, "2022-06-22T16:42:55-05:00"), (LCLOG_RESPONSE_NO_MESSAGE, "1970-01-01T00:00:00-00:00")),
    )
    def test_last_reboot(self, response, reboot_time):
        """Return the last reboot time."""
        reboot_time = datetime.fromisoformat(reboot_time)
        self.requests_mock.get("/redfish/v1/Managers/iDRAC.Embedded.1/Logs/Lclog", json=response)
        assert self.redfish.last_reboot() == reboot_time

    @pytest.mark.parametrize("generation", (1, 13, 14))
    @mock.patch("spicerack.redfish.time.sleep")
    @mock.patch("spicerack.redfish.Redfish.last_reboot")
    @mock.patch("spicerack.redfish.Redfish.check_connection")
    def test_wait_reboot_since(self, mocked_check_connection, mocked_last_reboot, mocked_sleep, generation):
        """It should return immediately if the host has already rebooted."""
        since = datetime.fromisoformat("2022-01-01T00:00:00-00:00")
        mocked_check_connection.return_value = True
        mocked_last_reboot.return_value = datetime.fromisoformat("2022-01-01T00:05:00-00:00")
        self.redfish._generation = generation  # pylint: disable=protected-access
        self.redfish.wait_reboot_since(since)
        mocked_check_connection.called_once()
        mocked_last_reboot.called_once_with(since)
        if generation < 14:
            mocked_sleep.called_once_with(120)
        mocked_sleep.called_once_with(30)

    @pytest.mark.parametrize("generation", (1, 13, 14))
    @mock.patch("spicerack.redfish.time.sleep")
    @mock.patch("spicerack.redfish.Redfish.last_reboot")
    @mock.patch("spicerack.redfish.Redfish.check_connection")
    def test_wait_reboot_since_to_early(self, mocked_check_connection, mocked_last_reboot, mocked_sleep, generation):
        """It should raise an error if the reboot time is to early."""
        since = datetime.fromisoformat("2022-01-01T00:05:00-00:00")
        mocked_check_connection.return_value = True
        mocked_last_reboot.return_value = datetime.fromisoformat("2022-01-01T00:00:00-00:00")
        self.redfish._generation = generation  # pylint: disable=protected-access
        with pytest.raises(redfish.RedfishError, match="no new reboot detected"):
            self.redfish.wait_reboot_since(since)
        mocked_check_connection.called_once()
        mocked_last_reboot.called_once_with(since)
        if generation < 14:
            mocked_sleep.called_once_with(120)

    @pytest.mark.parametrize("method", ("get", "head"))
    def test_request_dry_run_ro(self, method):
        """It should perform any RO request and return the actual response also in dry_run mode."""
        self.requests_mock.get("/redfish", json={"v1": "/redfish/v1/"})
        self.requests_mock.head("/redfish")
        response = self.redfish_dry_run.request(method, "/redfish")
        assert response.status_code == 200
        if method == "get":
            assert response.json() == {"v1": "/redfish/v1/"}

    @pytest.mark.parametrize("method", ("connect", "delete", "options", "patch", "post", "put", "trace"))
    def test_request_dry_run_rw(self, method):
        """It should not perform any RW request and return a dummy successful response in dry-run mode."""
        response = self.redfish_dry_run.request(method, "/redfish")
        assert response.status_code == 200
        assert not response.text

    def test_request_dry_run_fail(self, caplog):
        """If the request fails in dry-run mode, it should return a dummy successful response."""
        self.requests_mock.get("/redfish", exc=requests.exceptions.ConnectTimeout)
        with caplog.at_level(logging.ERROR):
            response = self.redfish_dry_run.request("get", "/redfish")

        assert response.status_code == 200
        assert "Failed to perform GET request to https://test.example.org/redfish" in caplog.text

    def test_request_ok(self):
        """It should perform the provided request and return it."""
        self.requests_mock.get("/redfish", json={"v1": "/redfish/v1/"})
        response = self.redfish.request("get", "/redfish")
        assert response.json() == {"v1": "/redfish/v1/"}
        assert response.status_code == 200

    def test_request_response_wrong_status_code(self):
        """It should raise a RedfishError if the request returns an error status code."""
        self.requests_mock.post("/redfish", json={"error": {"code": "1.0", "message": "error"}}, status_code=405)
        with pytest.raises(
            redfish.RedfishError, match="POST https://test.example.org/redfish returned HTTP 405 with message"
        ):
            self.redfish.request("post", "/redfish", json={"key": "value"})

    def test_request_response_raises(self):
        """It should raise a RedfishError if the request failes to be performed."""
        self.requests_mock.get("/redfish", exc=requests.exceptions.ConnectTimeout)
        with pytest.raises(
            redfish.RedfishError, match="Failed to perform GET request to https://test.example.org/redfish"
        ):
            self.redfish.request("get", "/redfish")

    def test_request_invalid_uri(self):
        """It should raise a RedfishError if the URI is invalid."""
        with pytest.raises(redfish.RedfishError, match="Invalid uri redfish, it must start with a /"):
            self.redfish.request("get", "redfish")

    def test_submit_task_dry_run(self):
        """In dry-run mode should not submit a task and return a dummy location."""
        assert self.redfish_dry_run.submit_task("/redfish/v1/SomeAction") == "/"

    def test_submit_task_ok(self):
        """It should submit the request and return the URI for polling the task results."""
        self.requests_mock.post(
            "/redfish/v1/SomeAction",
            status_code=202,
            headers={"Location": "/redfish/v1/TaskService/Tasks/JID_1234567890"},
        )
        response = self.redfish.submit_task("/redfish/v1/SomeAction")
        assert response == "/redfish/v1/TaskService/Tasks/JID_1234567890"

    def test_submit_task_invalid_code(self):
        """It should raise a RedfishError if the status code is not 202."""
        self.requests_mock.post(
            "/redfish/v1/SomeAction",
            headers={"Location": "/redfish/v1/TaskService/Tasks/JID_1234567890"},
        )
        with pytest.raises(redfish.RedfishError, match="expected HTTP 202, got HTTP 200 instead"):
            self.redfish.submit_task("/redfish/v1/SomeAction")

    def test_submit_task_no_location(self):
        """It should raise a RedfishError if there is no Location header in the response."""
        self.requests_mock.post("/redfish/v1/SomeAction", status_code=202)
        with pytest.raises(redfish.RedfishError, match="Unable to get the task URI to poll results"):
            self.redfish.submit_task("/redfish/v1/SomeAction")

    def test_connection_ok(self):
        """It should not raise if able to connect to the Redfish API."""
        self.requests_mock.get("/redfish", json={"v1": "/redfish/v1/"})
        self.redfish.check_connection()

    def test_connection_fail(self):
        """It should raise a RedfishError if unable to connect to the Redfish API."""
        self.requests_mock.get("/redfish", status_code=400)
        with pytest.raises(redfish.RedfishError, match="GET https://test.example.org/redfish returned HTTP 400"):
            self.redfish.check_connection()

    def test_poll_task_dry_run(self):
        """It should return a dummy response in dry-run mode."""
        assert self.redfish_dry_run.poll_task("/redfish/v1/TaskService/Tasks/JID_1234567890") == {}

    @pytest.mark.parametrize("response", (DELL_TASK_REPONSE, DELL_TASK_REPONSE_EPOC, DELL_TASK_REPONSE_BAD_TIME))
    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_poll_endtime(self, mocked_sleep, response):
        """It should raise a RedfishError if polling the task the device returns a failure code."""
        self.requests_mock.get("/redfish/v1/TaskService/Tasks/JID_1234567890", status_code=202, json=response)
        with pytest.raises(redfish.RedfishTaskNotCompletedError):
            self.redfish.poll_task("/redfish/v1/TaskService/Tasks/JID_1234567890")
        mocked_sleep.assert_called()

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_poll_task_raises(self, mocked_sleep):
        """It should raise a RedfishError if polling the task the device returns a failure code."""
        self.requests_mock.get("/redfish/v1/TaskService/Tasks/JID_1234567890", status_code=201)
        with pytest.raises(redfish.RedfishError, match="JID_1234567890 returned HTTP 201"):
            self.redfish.poll_task("/redfish/v1/TaskService/Tasks/JID_1234567890")

        mocked_sleep.assert_not_called()

    def test_find_account_ok(self):
        """It should return the URI of the account with the given username."""
        add_accounts_mock_responses(self.requests_mock)
        uri, etag = self.redfish.find_account("root")
        assert uri == "/redfish/v1/AccountService/Accounts/2"
        assert etag == "12345-2"

    def test_find_account_raises(self):
        """It should raise a RedfishError if the user is not found."""
        add_accounts_mock_responses(self.requests_mock)
        with pytest.raises(redfish.RedfishError, match="Unable to find account for username nonexistent"):
            self.redfish.find_account("nonexistent")

    @pytest.mark.parametrize("user_id, username, is_current", ((1, "user", False), (2, "root", True)))
    def test_change_user_password_ok(self, user_id, username, is_current):
        """It should change the password for the given user and update the instance auth credentials accordingly."""
        self.requests_mock.get("/redfish", json={"v1": "/redfish/v1/"})
        add_accounts_mock_responses(self.requests_mock)
        response = {
            "@Message.ExtendedInfo": [
                {"Message": "Changed password", "MessageId": "123", "Severity": "Informational", "Resolution": "None"}
            ]
        }
        self.requests_mock.patch(f"/redfish/v1/AccountService/Accounts/{user_id}", json=response)

        current_auth = self.redfish.request("get", "/redfish").request.headers["Authorization"]
        self.redfish.change_user_password(username, "test1234")
        new_auth = self.redfish.request("get", "/redfish").request.headers["Authorization"]

        if is_current:  # Check that the authorization header changed
            assert new_auth != current_auth
        else:
            assert new_auth == current_auth

    def test_change_user_password_raises(self):
        """It should raise a RedfishError if the reponse is not what HTTP 200."""
        add_accounts_mock_responses(self.requests_mock)
        self.requests_mock.patch("/redfish/v1/AccountService/Accounts/2", status_code=202)
        with pytest.raises(redfish.RedfishError, match="Got unexpected HTTP 202, expected 200"):
            self.redfish.change_user_password("root", "test1234")

    def test_get_power_state(self):
        """It should return the current power state of the device."""
        self.requests_mock.get("/redfish/v1/Chassis/System.Embedded.1", json={"PowerState": "On"})
        assert self.redfish.get_power_state() == "On"

    @pytest.mark.parametrize("action", tuple(redfish.ChassisResetPolicy))
    def test_chassis_reset_ok(self, action):
        """It should perform a chassis reset with the given action."""
        self.requests_mock.post("/redfish/v1/Systems/System.Embedded.1/Actions/ComputerSystem.Reset", status_code=204)
        self.redfish.chassis_reset(action)

    def test_chassis_reset_raises(self):
        """It should raise a RedfishError if the response code of the chassis reset operation is not 204."""
        self.requests_mock.post("/redfish/v1/Systems/System.Embedded.1/Actions/ComputerSystem.Reset")
        with pytest.raises(redfish.RedfishError, match="Got unexpected response HTTP 200, expected HTTP 204"):
            self.redfish.chassis_reset(redfish.ChassisResetPolicy.FORCE_OFF)


class TestDellSCP:
    """Tests for the DellSCP class."""

    def setup_method(self):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        self.config = redfish.DellSCP(deepcopy(DELL_SCP), redfish.DellSCPTargetPolicy.ALL)
        self.config_allow_new = redfish.DellSCP(
            deepcopy(DELL_SCP), redfish.DellSCPTargetPolicy.ALL, allow_new_attributes=True
        )

    @pytest.mark.parametrize(
        "property_name, expected",
        (
            ("config", DELL_SCP),
            ("target", redfish.DellSCPTargetPolicy.ALL),
            ("service_tag", "12ABC34"),
            ("model", "PowerEdge R440"),
            ("timestamp", datetime(2021, 12, 9, 9, 32, 6)),
            ("comments", ["First comment"]),
            (
                "components",
                {
                    "Some.Component.1": {"Some.Attribute.1": "value"},
                    "Some.Component.2": {"Some.Attribute.1": "value", "Some.Attribute.2": "value"},
                    "List.Component.1": {
                        "Comma.Separated.List.1": "value1,value2",
                        "Comma.Space.Separated.List.1": "value1, value2",
                    },
                },
            ),
        ),
    )
    def test_properties(self, property_name, expected):
        """It should return the value of the given property."""
        assert getattr(self.config, property_name) == expected
        assert getattr(self.config_allow_new, property_name) == expected

    @pytest.mark.parametrize(
        "component, attribute, exception_message",
        (
            ("Non.Existent", "Some.Attribute.1", "Unable to find component Non.Existent"),
            ("Some.Component.1", "Non.Existent", "Unable to find attribute Some.Component.1 -> Non.Existent"),
        ),
    )
    def test_set_raise(self, component, attribute, exception_message):
        """It should raise RedfishError if the specified component or attribute does not exists and is now allowed."""
        with pytest.raises(redfish.RedfishError, match=exception_message):
            self.config.set(component, attribute, "new_value")

    def test_set_new_attribute(self):
        """It should add the new attribute if allow_new_attributes is set to True or empty_components() was called."""
        params = ["Some.Component.1", "Non.Existent", "new_value"]
        self.config_allow_new.set(*params)
        assert self.config_allow_new.components[params[0]][params[1]] == params[2]
        self.config.empty_components()
        self.config.set(*params)
        assert self.config.components[params[0]][params[1]] == params[2]

    def test_set_raise_new_component(self):
        """It should raise RedfishError if the component does not exists even if allow_new_attributes is set."""
        with pytest.raises(redfish.RedfishError, match="Unable to find component Non.Existent"):
            self.config_allow_new.set("Non.Existent", "Some.Attribute.1", "new_value")

    def test_set_new_component(self):
        """It should create a new component and add to it the new attribute if empty_components() was called."""
        self.config.empty_components()
        self.config.set("Non.Existent", "Some.Attribute.1", "new_value")
        assert self.config.components["Non.Existent"]["Some.Attribute.1"] == "new_value"

    @pytest.mark.parametrize(
        "index, component, attribute, value",
        (
            (1, "Some.Component.1", "Some.Attribute.1", "value"),
            (2, "List.Component.1", "Comma.Separated.List.1", "value1,value2"),
            (2, "List.Component.1", "Comma.Space.Separated.List.1", "value1, value2"),
        ),
    )
    def test_set_same_value(self, index, component, attribute, value, caplog):
        """It should not set the same value and log a different message if the value is already correct."""
        with caplog.at_level(logging.INFO):
            was_changed = self.config.set(component, attribute, value)

        assert not was_changed
        assert f"Skipped set of attribute {component} -> {attribute}, has already" in caplog.text
        assert all(
            i["Set On Import"] == "False"
            for i in self.config.config["SystemConfiguration"]["Components"][index]["Attributes"]
        )

    def test_set_new_value(self):
        """It should set the value and mark it for import."""
        was_changed = self.config.set("Some.Component.1", "Some.Attribute.1", "new value")
        attribute = self.config.config["SystemConfiguration"]["Components"][1]["Attributes"][0]
        assert was_changed
        assert attribute["Value"] == "new value"
        assert attribute["Set On Import"] == "True"

    def test_update(self):
        """It should update the config with the given changes."""
        changes = {
            "Some.Component.1": {"Some.Attribute.1": "value"},
            "Some.Component.2": {"Some.Attribute.2": "new value"},
        }
        was_changed = self.config.update(changes)
        assert was_changed
        assert self.config.components["Some.Component.1"]["Some.Attribute.1"] == "value"
        assert self.config.components["Some.Component.2"]["Some.Attribute.2"] == "new value"

    def test_update_no_changes(self):
        """It should return False if no changes were made."""
        changes = {
            "Some.Component.1": {"Some.Attribute.1": "value"},
            "Some.Component.2": {"Some.Attribute.2": "value"},
        }
        was_changed = self.config.update(changes)
        assert not was_changed
        assert self.config.components["Some.Component.1"]["Some.Attribute.1"] == "value"
        assert self.config.components["Some.Component.2"]["Some.Attribute.2"] == "value"

    def test_empty_components(self):
        """It should empty the components of the current configuration."""
        self.config.empty_components()
        assert self.config.components == {}


class TestRedfishDell:
    """Tests for the RedfishDell class."""

    @pytest.fixture(autouse=True)
    def setup_method(self, requests_mock):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        self.redfish = redfish.RedfishDell("test.example.org", "root", "mysecret", dry_run=False)
        self.requests_mock = requests_mock

    @pytest.mark.parametrize("allow_new", (False, True))
    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_scp_dump(self, mocked_sleep, allow_new):
        """It should return an instance of DellSCP with the current configuration for the given target."""
        self.requests_mock.post(
            "/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Oem/EID_674_Manager.ExportSystemConfiguration",
            headers={"Location": "/redfish/v1/TaskService/Tasks/JID_1234567890"},
            status_code=202,
        )
        self.requests_mock.get(
            "/redfish/v1/TaskService/Tasks/JID_1234567890",
            [{"status_code": 202, "json": DELL_TASK_REPONSE}, {"status_code": 200, "json": DELL_SCP}],
        )
        config = self.redfish.scp_dump(allow_new_attributes=allow_new)
        assert config.service_tag == "12ABC34"
        assert mocked_sleep.called
        if allow_new:
            config.set("Some.Component.1", "Non.Existent", "new_value")
            assert config.components["Some.Component.1"]["Non.Existent"] == "new_value"
        else:
            with pytest.raises(redfish.RedfishError, match="Unable to find attribute Some.Component.1 -> Non.Existent"):
                config.set("Some.Component.1", "Non.Existent", "new_value")

    @pytest.mark.parametrize(
        "uri_suffix, preview",
        (
            ("ImportSystemConfigurationPreview", True),
            ("ImportSystemConfiguration", False),
        ),
    )
    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_scp_push(self, mocked_sleep, uri_suffix, preview):
        """It should push the configuration to the device for preview, no changes will be applied."""
        expected = deepcopy(DELL_TASK_REPONSE)
        expected["EndTime"] = "2021-12-09T14:39:29-06:00"
        self.requests_mock.post(
            f"/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Oem/EID_674_Manager.{uri_suffix}",
            headers={"Location": "/redfish/v1/TaskService/Tasks/JID_1234567890"},
            status_code=202,
        )
        self.requests_mock.get(
            "/redfish/v1/TaskService/Tasks/JID_1234567890",
            [{"status_code": 202, "json": DELL_TASK_REPONSE}, {"status_code": 200, "json": expected}],
        )
        result = self.redfish.scp_push(redfish.DellSCP(DELL_SCP, redfish.DellSCPTargetPolicy.ALL), preview=preview)
        assert result == expected
        assert mocked_sleep.called



"""ElasticsearchCluster module."""
import logging
from collections import defaultdict
from contextlib import ExitStack, contextmanager
from datetime import datetime, timedelta
from math import floor
from random import shuffle
from typing import DefaultDict, Dict, Iterable, Iterator, List, Optional, Sequence, Set

import curator
from elasticsearch import ConflictError, Elasticsearch, RequestError, TransportError
from urllib3.exceptions import HTTPError
from wmflib.prometheus import Prometheus

from spicerack.administrative import Reason
from spicerack.decorators import retry
from spicerack.exceptions import SpicerackCheckError, SpicerackError
from spicerack.remote import Remote, RemoteHosts, RemoteHostsAdapter

logger = logging.getLogger(__name__)


class ElasticsearchClusterError(SpicerackError):
    """Custom Exception class for errors of this module."""


class ElasticsearchClusterCheckError(SpicerackCheckError):
    """Custom Exception class for check errors of this module."""


def create_elasticsearch_clusters(
    configuration: Dict[str, Dict[str, Dict[str, str]]],
    clustergroup: str,
    write_queue_datacenters: Sequence[str],
    remote: Remote,
    prometheus: Prometheus,
    dry_run: bool = True,
) -> "ElasticsearchClusters":
    """Create ElasticsearchClusters instance.

    Arguments:
        clustergroup (str): name of cluster group.
        write_queue_datacenters (Sequence[str]): Sequence of which core DCs to query write queues for.
        remote (spicerack.remote.Remote): the Remote instance.
        prometheus (wmflib.prometheus.Prometheus): the prometheus instance.
        dry_run (bool, optional):  whether this is a DRY-RUN.

    Raises:
        spicerack.elasticsearch_cluster.ElasticsearchClusterError: Thrown when the requested cluster configuration is
            not found.

    Returns:
        spicerack.elasticsearch_cluster.ElasticsearchClusters: ElasticsearchClusters instance.

    """
    try:
        endpoints = configuration["search"][clustergroup].values()
    except KeyError as e:
        raise ElasticsearchClusterError(f"No cluster group named {clustergroup}") from e

    clusters = [Elasticsearch(endpoint) for endpoint in endpoints]
    elasticsearch_clusters = [ElasticsearchCluster(cluster, remote, dry_run=dry_run) for cluster in clusters]
    return ElasticsearchClusters(
        elasticsearch_clusters,
        remote,
        prometheus,
        write_queue_datacenters,
        dry_run=dry_run,
    )


class ElasticsearchHosts(RemoteHostsAdapter):
    """Remotehosts Adapter for managing elasticsearch nodes."""

    def __init__(
        self,
        remote_hosts: RemoteHosts,
        nodes: Sequence["NodesGroup"],
        dry_run: bool = True,
    ) -> None:
        """After calling the super's constructor, initialize other instance variables.

        Arguments:
            remote_hosts (spicerack.remote.RemoteHosts): the instance with the target hosts.
            nodes (list): list of dicts containing clusters hosts belong to.
            dry_run (bool, optional): whether this is a DRY-RUN.

        """
        super().__init__(remote_hosts)
        self._nodes = nodes
        self._dry_run = dry_run

    def get_remote_hosts(self) -> RemoteHosts:
        """Returns elasticsearch remote hosts.

        Returns:
            spicerack.remote.RemoteHosts: RemoteHosts instance for this adapter.

        """
        return self._remote_hosts

    def start_elasticsearch(self) -> None:
        """Starts all elasticsearch instances."""
        self._systemctl_for_each_instance("start")

    def stop_elasticsearch(self) -> None:
        """Stops all elasticsearch instances."""
        self._systemctl_for_each_instance("stop")

    def restart_elasticsearch(self) -> None:
        """Restarts all elasticsearch instances."""
        self._systemctl_for_each_instance("restart")

    def _systemctl_for_each_instance(self, action: str) -> None:
        logger.info("%s all elasticsearch instances on %s", action, self)
        self._remote_hosts.run_sync(f"cat /etc/elasticsearch/instances | xargs systemctl {action}")

    def depool_nodes(self) -> None:
        """Depool the hosts."""
        logger.info("Depooling %s", self)
        self._remote_hosts.run_sync("depool")

    def pool_nodes(self) -> None:
        """Pool the hosts."""
        logger.info("Pooling %s", self)
        self._remote_hosts.run_sync("pool")

    def wait_for_elasticsearch_up(self, timeout: timedelta = timedelta(minutes=15)) -> None:
        """Check if elasticsearch instances on each node are up.

        Arguments:
            timeout (datetime.timedelta, optional): represent how long to wait for all instances to be up.

        """
        delay = timedelta(seconds=5)
        tries = max(floor(timeout / delay), 1)

        logger.info("waiting for elasticsearch instances to come up on %s", self)

        @retry(
            tries=tries,
            delay=delay,
            backoff_mode="constant",
            exceptions=(ElasticsearchClusterError, ElasticsearchClusterCheckError),
        )
        def inner_wait() -> None:
            for node in self._nodes:
                node.check_all_nodes_up()

        if not self._dry_run:
            inner_wait()


class ElasticsearchClusters:
    """Class to manage elasticsearch clusters."""

    def __init__(
        self,
        clusters: Sequence["ElasticsearchCluster"],
        remote: Remote,
        prometheus: Prometheus,
        write_queue_datacenters: Sequence[str],
        dry_run: bool = True,
    ) -> None:
        """Initialize ElasticsearchClusters.

        Arguments:
            clusters (list): list of :py:class:`spicerack.elasticsearch_cluster.ElasticsearchCluster` instances.
            remote (spicerack.remote.Remote): the Remote instance.
            prometheus (wmflib.prometheus.Prometheus): the prometheus instance.
            write_queue_datacenters (Sequence[str]): Sequence of which core DCs to query write queues for.
            dry_run (bool, optional): whether this is a DRY-RUN.

        """
        self._clusters = clusters
        self._remote = remote
        self._prometheus = prometheus
        self._write_queue_datacenters = write_queue_datacenters
        self._dry_run = dry_run

    def __str__(self) -> str:
        """Class string method."""
        return str(self._clusters)

    def flush_markers(self, timeout: timedelta = timedelta(seconds=60)) -> None:
        """Flush markers on all clusters.

        Arguments:
            timeout (datetime.timedelta, optional): timedelta object for elasticsearch request timeout.

        """
        for cluster in self._clusters:
            cluster.flush_markers(timeout)

    def force_allocation_of_all_unassigned_shards(self) -> None:
        """Force allocation of unassigned shards on all clusters."""
        for cluster in self._clusters:
            cluster.force_allocation_of_all_unassigned_shards()

    @contextmanager
    def frozen_writes(self, reason: Reason) -> Iterator[List[None]]:
        """Freeze all writes to the clusters and then perform operations before unfreezing writes.

        Arguments:
            reason (spicerack.administrative.Reason): Reason for freezing writes.

        Yields:
            list: a side-effect list of :py:data:`None`, as a result of the stack of context managers.

        """
        logger.info("Freezing writes on %s", self)
        with ExitStack() as stack:
            yield [stack.enter_context(cluster.frozen_writes(reason)) for cluster in self._clusters]

    @contextmanager
    def stopped_replication(self) -> Iterator[List[None]]:
        """Stops replication for all clusters.

        Yields:
            list: a side-effect list of :py:data:`None`, as a result of the stack of context managers.

        """
        logger.info("stopping replication on %s", self)
        with ExitStack() as stack:
            yield [stack.enter_context(cluster.stopped_replication()) for cluster in self._clusters]

    def wait_for_green(self, timeout: timedelta = timedelta(hours=1)) -> None:
        """Wait for green on all clusters.

        Arguments:
            timeout (datetime.timedelta, optional): timedelta object to represent how long to wait for green status
                on all clusters.

        """
        delay = timedelta(seconds=10)
        tries = max(floor(timeout / delay), 1)
        logger.info("waiting for clusters to be green")

        @retry(
            tries=tries,
            delay=delay,
            backoff_mode="constant",
            exceptions=(ElasticsearchClusterCheckError,),
        )
        def inner_wait() -> None:
            for cluster in self._clusters:
                cluster.check_green()

        inner_wait()

    def wait_for_yellow_w_no_moving_shards(self, timeout: timedelta = timedelta(hours=1)) -> None:
        """Wait for a yellow cluster status with no relocating or initializing shards.

        Arguments:
            timeout (datetime.timedelta, optional): timedelta object to represent how long to wait
                 for no yellow status with no initializing or relocating shards on all clusters.

        """
        delay = timedelta(seconds=10)
        tries = max(floor(timeout / delay), 1)
        logger.info("waiting for clusters to be yellow with no initializing or relocating shards")

        @retry(
            tries=tries,
            delay=delay,
            backoff_mode="constant",
            exceptions=(ElasticsearchClusterCheckError,),
        )
        def inner_wait() -> None:
            for cluster in self._clusters:
                cluster.check_yellow_w_no_moving_shards()

        inner_wait()

    def get_next_clusters_nodes(self, started_before: datetime, size: int = 1) -> Optional[ElasticsearchHosts]:
        """Get next set of cluster nodes for cookbook operations like upgrade, rolling restart etc.

        Nodes are selected from the row with the least restarted nodes. This ensures that a row is fully upgraded
        before moving to the next row. Since shards cannot move to a node with an older version of elasticsearch,
        this should help to keep all shards allocated at all times.

        Master capable nodes are returned after all other nodes have restarted to support version upgrades which
        strongly suggest the masters are upgraded last.

        Arguments:
            started_before (datetime.datetime): the time against after which we check if the node has been restarted.
            size (int, optional): size of nodes not restarted in a row.

        Returns:
            spicerack.elasticsearch_cluster.ElasticsearchHosts: next eligible nodes for ElasticsearchHosts or
                :py:data:`None` when all nodes have been processed.

        """
        if size < 1:
            raise ElasticsearchClusterError("Size of next nodes must be at least 1")

        nodes_group = self._get_nodes_group()
        nodes_to_process = [node for node in nodes_group if not node.restarted_since(started_before)]
        if not nodes_to_process:
            return None
        # delay master nodes until all the workers have been restarted
        nodes_to_process = _restartable_node_groups(nodes_to_process)
        rows = ElasticsearchClusters._to_rows(nodes_to_process)
        sorted_rows = sorted(rows.values(), key=len)
        next_nodes = sorted_rows[0]
        next_nodes = next_nodes[:size]
        node_names = ",".join([node.fqdn for node in next_nodes])
        return ElasticsearchHosts(self._remote.query(node_names), next_nodes, dry_run=self._dry_run)

    def _get_nodes_group(self) -> Iterable["NodesGroup"]:
        """Create nodes_group for each nodes.

        Returns:
            dict: merged clusters nodes.

        """
        nodes_group: Dict[str, NodesGroup] = {}
        for cluster in self._clusters:
            for json_node in cluster.get_nodes().values():
                node_name = json_node["attributes"]["hostname"]
                if node_name not in nodes_group:
                    nodes_group[node_name] = NodesGroup(json_node, cluster)
                else:
                    nodes_group[node_name].accumulate(json_node, cluster)
        return nodes_group.values()

    @staticmethod
    def _to_rows(nodes: Sequence["NodesGroup"]) -> DefaultDict[str, List["NodesGroup"]]:
        """Arrange nodes in rows, so each node belongs in their respective row.

        Arguments:
            nodes (list): list containing dicts of elasticsearch nodes.

        Returns:
            defaultdict: defaultdict object containing a normalized rows of elasticsearch nodes. For example::

                {'row1': [{'name': 'el1'}, {'name': 'el2'}], 'row2': [{'name': 'el6'}]}

        """
        rows: DefaultDict[str, List[NodesGroup]] = defaultdict(list)
        for node in nodes:
            rows[node.row].append(node)
        return rows

    def reset_indices_to_read_write(self) -> None:
        """Reset all readonly indices to read/write.

        In some cases (running low on disk space), indices are switched to
        readonly. This method will update all readonly indices to read/write.
        """
        for cluster in self._clusters:
            cluster.reset_indices_to_read_write()

    @retry(
        tries=60,
        delay=timedelta(seconds=60),
        backoff_mode="constant",
        exceptions=(ElasticsearchClusterCheckError,),
    )
    def wait_for_all_write_queues_empty(self) -> None:
        """Wait for all relevant CirrusSearch write queues to be empty.

        Checks the Prometheus server in each of the CORE_DATACENTERS

        At most waits for 60*60 seconds = 1 hour.

        Does not retry if prometheus returns empty results for all datacenters.
        """
        # We expect all DCs except one to return empty results, but we have a problem if all return empty
        have_received_results = False

        for dc in self._write_queue_datacenters:
            query = (
                "kafka_burrow_partition_lag{"
                '    group="cpjobqueue-cirrusSearchElasticaWrite",'
                '    topic=~"[[:alpha:]]*.cpjobqueue.partitioned.mediawiki.job.cirrusSearchElasticaWrite"'
                "}"
            )
            # Query returns a list of dictionaries each of format {'metric': {}, 'value': [$timestamp, $value]}
            results = self._prometheus.query(query, dc)
            if not results:
                logger.info("Prometheus returned no results for query %s in dc %s", query, dc)
                continue

            have_received_results = True

            # queue_results => (topic, partition, value)
            queue_results = [
                (
                    partitioned_result["metric"]["topic"],
                    partitioned_result["metric"]["partition"],
                    int(partitioned_result["value"][1]),
                )
                for partitioned_result in results
            ]
            logger.debug("Prom query %s returned queue_results of %s", query, queue_results)

            # If any of the partitions are non-empty, raise an error
            for (topic, partition, queue_size) in queue_results:
                if queue_size > 0:
                    raise ElasticsearchClusterCheckError(
                        f"Write queue not empty (had value of {queue_size}) for partition {partition} of topic {topic}."
                    )
        if not have_received_results:
            raise ElasticsearchClusterError(
                f"Prometheus query {query} returned empty response for all dcs in {self._write_queue_datacenters}, "
                f"is query correct?"
            )


class ElasticsearchCluster:
    """Class to manage elasticsearch cluster."""

    def __init__(self, elasticsearch: Elasticsearch, remote: Remote, dry_run: bool = True) -> None:
        """Initialize ElasticsearchCluster.

        Arguments:
            elasticsearch (elasticsearch.Elasticsearch): elasticsearch instance.
            remote (spicerack.remote.Remote): the Remote instance.
            dry_run (bool, optional):  whether this is a DRY-RUN.

        """
        self._elasticsearch = elasticsearch
        self._remote = remote
        self._dry_run = dry_run
        self._freeze_writes_index: str = "mw_cirrus_metastore"
        self._freeze_writes_doc_type: str = "mw_cirrus_metastore"

    def __str__(self) -> str:
        """Class string method."""
        return str(self._elasticsearch)

    def get_nodes(self) -> Dict:
        """Get all Elasticsearch Nodes.

        Returns:
            dict: dictionary of elasticsearch nodes in the cluster.

        """
        try:
            return self._elasticsearch.nodes.info()["nodes"]
        except (TransportError, HTTPError) as e:
            raise ElasticsearchClusterError("Could not connect to the cluster") from e

    def is_node_in_cluster_nodes(self, node: str) -> bool:
        """Checks if node is in a list of elasticsearch cluster nodes.

        Arguments:
            node (str): the elasticsearch host.

        Returns:
            bool: :py:data:`True` if node is present and :py:data:`False` if not.

        """
        nodes_names = [node["attributes"]["hostname"] for node in self.get_nodes().values()]
        if node in nodes_names:
            return True

        return False

    @contextmanager
    def stopped_replication(self) -> Iterator[None]:
        """Context manager to perform actions while the cluster replication is stopped."""
        self._stop_replication()
        try:
            yield
        finally:
            self._start_replication()

    def _stop_replication(self) -> None:
        """Stops cluster replication."""
        logger.info("stop replication - %s", self)
        self._do_cluster_routing(
            curator.ClusterRouting(
                self._elasticsearch,
                routing_type="allocation",
                setting="enable",
                value="primaries",
                wait_for_completion=False,
            )
        )

    def _start_replication(self) -> None:
        """Starts cluster replication."""
        logger.info("start replication - %s", self)
        self._do_cluster_routing(
            curator.ClusterRouting(
                self._elasticsearch,
                routing_type="allocation",
                setting="enable",
                value="all",
                wait_for_completion=False,
            )
        )

    def _do_cluster_routing(self, cluster_routing: curator.ClusterRouting) -> None:
        """Performs cluster routing of shards.

        Arguments:
            cluster_routing (curator.ClusterRouting): Curator's cluster routing object.

        """
        if self._dry_run:
            cluster_routing.do_dry_run()
        else:
            cluster_routing.do_action()

    def check_green(self) -> None:
        """Cluster health status.

        Raises:
            spicerack.elasticsearch_cluster.ElasticsearchClusterCheckError:
                This is raised when request times and cluster is not green.

        """
        try:
            self._elasticsearch.cluster.health(wait_for_status="green", timeout="1s")
        except (TransportError, HTTPError) as e:
            raise ElasticsearchClusterCheckError("Error while waiting for green") from e

    def check_yellow_w_no_moving_shards(self) -> None:
        """Cluster health status.

        Raises:
            spicerack.elasticsearch_cluster.ElasticsearchClusterCheckError:
                This is raised when request times and cluster is not yellow with no initializing or relocating shards.

        """
        try:
            self._elasticsearch.cluster.health(
                wait_for_status="yellow",
                wait_for_no_initializing_shards=True,
                wait_for_no_relocating_shards=True,
                timeout="1s",
            )
        except (TransportError, HTTPError) as e:
            raise ElasticsearchClusterCheckError(
                "Error while waiting for yellow with no initializing or relocating shards"
            ) from e

    @contextmanager
    def frozen_writes(self, reason: Reason) -> Iterator[None]:
        """Stop writes to all elasticsearch indices and enable them on exit.

        Arguments:
            reason (spicerack.administrative.Reason): Reason for freezing writes.

        """
        self._freeze_writes(reason)
        try:
            yield
        finally:
            try:
                self._unfreeze_writes()
            except ElasticsearchClusterError as e:
                # Unfreeze failed, we can try to freeze and unfreeze again,
                # which might work. If it throws an exception again, we won't
                # try a third time and let that new exception bubble up.
                logger.warning(
                    "Could not unfreeze writes, trying to freeze and unfreeze again: %s",
                    e,
                )
                self._freeze_writes(reason)
                self._unfreeze_writes()

    def _freeze_writes(self, reason: Reason) -> None:
        """Stop writes to all elasticsearch indices.

        Arguments:
            reason (spicerack.administrative.Reason): Reason for freezing writes.

        """
        doc = {
            "host": reason.hostname,
            "timestamp": datetime.utcnow().timestamp(),
            "reason": str(reason),
        }
        logger.info("Freezing all indices in %s", self)
        if self._dry_run:
            return
        try:
            self._elasticsearch.index(
                index=self._freeze_writes_index,
                doc_type=self._freeze_writes_doc_type,
                id="freeze-everything",
                body=doc,
            )
        except TransportError as e:
            raise ElasticsearchClusterError("Encountered error while creating document to freeze cluster writes") from e

    def _unfreeze_writes(self) -> None:
        """Enable writes on all elasticsearch indices."""
        logger.info("Unfreezing all indices in %s", self)
        if self._dry_run:
            return
        try:
            self._elasticsearch.delete(
                index=self._freeze_writes_index,
                doc_type=self._freeze_writes_doc_type,
                id="freeze-everything",
            )
        except TransportError as e:
            raise ElasticsearchClusterError(
                "Encountered error while deleting document to unfreeze cluster writes"
            ) from e

    def flush_markers(self, timeout: timedelta = timedelta(seconds=60)) -> None:
        """Flush markers unsynced.

        Note:
            ``flush`` and ``flush_synced`` are called here because from experience, it results in fewer shards not
            syncing. This also makes the recovery faster.

        Arguments:
            timeout (datetime.timedelta): timedelta object for elasticsearch request timeout.

        """
        logger.info("flush markers on %s", self)
        try:
            self._elasticsearch.indices.flush(force=True, request_timeout=timeout.seconds)
        except ConflictError:
            logger.warning("Not all shards were flushed on %s.", self)

        try:
            self._elasticsearch.indices.flush_synced(request_timeout=timeout.seconds)
        except ConflictError:
            logger.warning("Not all shards were synced flushed on %s.", self)

    def force_allocation_of_all_unassigned_shards(self) -> None:
        """Manual allocation of unassigned shards."""
        cluster_nodes_names = [node["name"] for node in self.get_nodes().values()]
        unassigned_shards = self._get_unassigned_shards()
        for unassigned_shard in unassigned_shards:
            self._force_allocation_of_shard(unassigned_shard, cluster_nodes_names)

    def _get_unassigned_shards(self) -> List[Dict]:
        """Fetch unassigned shards.

        Returns:
            list: list of unassigned shards from the cluster.

        """
        shards = self._elasticsearch.cat.shards(format="json", h="index,shard,state")
        return [s for s in shards if s["state"] == "UNASSIGNED"]

    def _force_allocation_of_shard(self, shard: Dict, nodes: List[str]) -> None:
        """Force allocation of shard.

        Arguments:
            shard (dict): shard of an index to be relocated.
            nodes (list): list of nodes to allocate shards to.

        Todo:
            It was found that forcing allocation of shards may perform better in terms of speed than
            letting elasticsearch do its recovery on its own.
            We should verify from time to time that elastic recovery performance has not gotten better
            and remove this step if proven unnecessary.

        """
        # shuffle nodes so that we don't allocate all shards on the same node
        shuffle(nodes)
        for node in nodes:
            try:
                logger.debug(
                    "Trying to allocate [%s:%s] on [%s]",
                    shard["index"],
                    shard["shard"],
                    node,
                )
                self._elasticsearch.cluster.reroute(
                    retry_failed=True,
                    body={
                        "commands": [
                            {
                                "allocate_replica": {
                                    "index": shard["index"],
                                    "shard": shard["shard"],
                                    "node": node,
                                }
                            }
                        ]
                    },
                )
                # successful allocation, we can exit
                logger.info(
                    "Successfully allocated shard [%s:%s] on [%s]",
                    shard["index"],
                    shard["shard"],
                    node,
                )
                break
            except RequestError:
                # error allocating shard, let's try the next node
                logger.debug(
                    "Could not reallocate shard [%s:%s] on [%s]",
                    shard["index"],
                    shard["shard"],
                    node,
                )
        else:
            logger.warning(
                "Could not reallocate shard [%s:%s] on any node",
                shard["index"],
                shard["shard"],
            )

    def reset_indices_to_read_write(self) -> None:
        """Reset all readonly indices to read/write.

        In some cases (running low on disk space), indices are switched to
        readonly. This method will update all readonly indices to read/write.
        """
        try:
            self._elasticsearch.indices.put_settings(body={"index.blocks.read_only_allow_delete": None}, index="_all")
        except (RequestError, TransportError, HTTPError) as e:
            raise ElasticsearchClusterError("Could not reset read only status") from e


class NodesGroup:
    """Internal class, used for parsing responses from the elasticsearch node API.

    Since the same server can host multiple elasticsearch instances, this class can consolidate those multiple
    instances in a single object.
    """

    def __init__(self, json_node: Dict, cluster: ElasticsearchCluster) -> None:
        """Instantiate a new node.

        Arguments:
            json_node (dict): a single node, as returned from the elasticsearch API.
            cluster (spicerack.elasticsearch_cluster.ElasticsearchCluster): an elasticsearch instance

        """
        self._hostname: str = json_node["attributes"]["hostname"]
        self._fqdn: str = json_node["attributes"]["fqdn"]
        cluster_name = json_node["settings"]["cluster"]["name"]
        self._clusters_names: List[str] = [cluster_name]
        self._clusters_instances: List[ElasticsearchCluster] = [cluster]
        self._row: str = json_node["attributes"]["row"]
        self._oldest_start_time = datetime.utcfromtimestamp(json_node["jvm"]["start_time_in_millis"] / 1000)
        self._master_capable: Set[str] = {cluster_name} if "master" in json_node["roles"] else set()

    def accumulate(self, json_node: Dict, cluster: ElasticsearchCluster) -> None:
        """Accumulate information from other elasticsearch instances running on the same server.

        Arguments:
            json_node (dict): a single node, as returned from the elasticsearch API.
            cluster (elasticsearch.Elasticsearch): an elasticsearch instance

        """
        if self._fqdn != json_node["attributes"]["fqdn"]:
            # should never happen
            fqdn1 = self._fqdn
            fqdn2 = json_node["attributes"]["fqdn"]
            raise AssertionError(f"Invalid data, two instances on the same node with different fqdns [{fqdn1}/{fqdn2}]")
        cluster_name = json_node["settings"]["cluster"]["name"]
        if cluster_name not in self._clusters_names:
            self._clusters_names.append(cluster_name)
        if cluster not in self._clusters_instances:
            self._clusters_instances.append(cluster)
        if self._row != json_node["attributes"]["row"]:
            # should never happen
            row1 = self._row
            row2 = json_node["attributes"]["row"]
            raise AssertionError(
                f"Invalid data, two instances on the same node with different rows {self._hostname}:[{row1}/{row2}]"
            )
        start_time = datetime.utcfromtimestamp(json_node["jvm"]["start_time_in_millis"] / 1000)
        self._oldest_start_time = min(self._oldest_start_time, start_time)
        if "master" in json_node["roles"]:
            self._master_capable.add(cluster_name)

    @property
    def row(self) -> str:
        """Datacenter row."""
        return self._row

    @property
    def fqdn(self) -> str:
        """Fully Qualified Domain Name."""
        return self._fqdn

    @property
    def clusters_names(self) -> Sequence[str]:
        """Names of cluster instances running on this node group."""
        return self._clusters_names

    @property
    def clusters_instances(self) -> Sequence[ElasticsearchCluster]:
        """Cluster instances running on this node group."""
        return self._clusters_instances

    @property
    def master_capable(self) -> Set[str]:
        """Set of clusters this node is master capable on."""
        return self._master_capable

    def restarted_since(self, since: datetime) -> bool:
        """Check if node has been restarted.

        Arguments:
            since (datetime.datetime): the time against after which we check if the node has been restarted.

        Returns:
            bool: True if the node has been restarted after since, false otherwise.

        """
        return self._oldest_start_time > since

    def check_all_nodes_up(self) -> None:
        """Check that all the nodes on this hosts are up and have joined their respective clusters.

        Raises:
            spicerack.elasticsearch_cluster.ElasticsearchClusterCheckError: if not all nodes have joined.

        """
        for cluster_instance in self._clusters_instances:
            if not cluster_instance.is_node_in_cluster_nodes(self._hostname):
                raise ElasticsearchClusterCheckError("Elasticsearch is not up yet")


def _restartable_node_groups(nodes: Sequence[NodesGroup]) -> List[NodesGroup]:
    """Returns the subset of nodes that are restartable.

    Primarily concerned with restarting nodes within a cluster prior to the
    masters of that cluster. Any node in the group that is master to another
    node in the group is removed.
    """
    masters = defaultdict(set)
    for node in nodes:
        for cluster_name in node.master_capable:
            masters[cluster_name].add(node.fqdn)
    rebootable = {node.fqdn for node in nodes}
    for node in nodes:
        for cluster_name in node.clusters_names:
            if cluster_name not in node.master_capable:
                # This node must be restarted before the masters
                rebootable = rebootable.difference(masters[cluster_name])
    if nodes and not rebootable:
        raise Exception("Restart dependency graph has cycles")
    return [node for node in nodes if node.fqdn in rebootable]



"""Provides methods to manipulate offsets set for specific consumer groups."""
import logging
import ssl
from dataclasses import dataclass
from datetime import timedelta
from typing import Any, Dict, List, Set

from kafka import KafkaConsumer, OffsetAndMetadata, TopicPartition

from spicerack.exceptions import SpicerackError

TIMEOUT_MS = 20000
"""int: Timeout set for any kafka operations used in this module."""

DELTA = timedelta(minutes=2).seconds * 1000
"""int: For offset approximation, timestamp that will be used will be earlier by this amount of ms."""

logger = logging.getLogger(__name__)


@dataclass
class ConsumerDefinition:
    """Data needed to identify a Kafka Consumer.

    Arguments:
        site (str): Kafka site/DC.
        cluster (str): Kafka cluster.
        consumer_group (str): Kafka consumer group.

    """

    site: str
    cluster: str
    consumer_group: str


class KafkaError(SpicerackError):
    """Custom exception class for errors in this module."""


class KafkaClient:
    """Class encapsulating Kafka operations for specific site, cluster and consumer group."""

    _consumer: KafkaConsumer
    _site: str

    def __init__(self, consumer_definition: ConsumerDefinition, kafka_config: Dict, dry_run: bool) -> None:
        """Sets up a KafkaConsumer.

        Arguments:
            consumer_definition (spicerack.kafka.ConsumerDefinition): Definition of the Kafka data for the consumer.
            kafka_config (dict): Complete, available in Puppet, kafka definition.
            dry_run (bool): Enable dry run mode.

        """
        self._dry_run = dry_run
        self._site = consumer_definition.site
        context = ssl.create_default_context()
        context.check_hostname = False  # TODO can be changed after https://phabricator.wikimedia.org/T291905
        crt_location = "/etc/ssl/certs/ca-certificates.crt"
        context.load_verify_locations(crt_location)

        logger.debug(
            "Creating kafka client (kafka consumer with site prefix) with SSL context (check_hostname: %s, "
            "verify_mode: %s) for cluster and site: %s, %s",
            context.check_hostname,
            context.verify_mode,
            consumer_definition.cluster,
            consumer_definition.site,
        )
        brokers = kafka_config[consumer_definition.cluster][consumer_definition.site]["brokers"]["ssl_string"]
        self._consumer = KafkaConsumer(
            bootstrap_servers=brokers,
            enable_auto_commit=False,
            consumer_timeout_ms=TIMEOUT_MS,
            request_timeout_ms=TIMEOUT_MS,
            group_id=consumer_definition.consumer_group,
            security_protocol="SSL",
            ssl_context=context,
        )

    def _get_full_topic_name(self, topic: str) -> str:
        """Construct site specific full topic name.

        Arguments:
            topic (str): Topic name without site prefix.

        Returns:
            str: Topic name with site prefix.

        """
        return f"{self._site}.{topic}"

    def get_committed_offset(self, topic_partition: TopicPartition) -> int:
        """Retrieve a committed offset for given TopicPartition.

        Arguments:
            topic_partition (kafka.structs.TopicPartition): Non-localized topic partition.

        Returns:
            int: Last committed offset.

        """
        localized_tp = self._get_localized_tp(topic_partition)
        committed_offset = self._consumer.committed(localized_tp)
        if committed_offset is None:
            raise KafkaError(f"Offset not found for topic {localized_tp.topic}, partition {localized_tp.partition}.")

        return committed_offset

    def _get_localized_tp(self, topic_partition: TopicPartition) -> TopicPartition:
        """Translate provided TopicPartition into the one local to the cluster.

        Arguments:
            topic_partition (kafka.structs.TopicPartition): Non-localized topic partition.

        Returns:
            kafka.structs.TopicPartition: topic and partition localized for current site.

        """
        return TopicPartition(self._get_full_topic_name(topic_partition.topic), topic_partition.partition)

    def get_next_timestamp(self, topic_partition: TopicPartition) -> int:
        """Retrieve a timestamp for given TopicPartition.

        Arguments:
            topic_partition (kafka.structs.TopicPartition): Non-localized topic partition.

        Returns:
            int: Currently about to be processed timestamp.

        """
        localized_tp = self._get_localized_tp(topic_partition)
        self._consumer.assign([localized_tp])
        msg = next(self._consumer, None)
        if msg is None:
            raise KafkaError(f"Offset not found for topic {localized_tp.topic}, partition {localized_tp.partition}.")
        return msg.timestamp

    def partitions_for_topic(self, topic_name: str) -> Set[int]:
        """Get partitions for a localized provided topic.

        Arguments:
            topic_name (str): Topic name without site prefix.

        Returns:
            set[int]: Set of partitions available for given topic.

        """
        full_topic_name = self._get_full_topic_name(topic_name)
        topic_partitions = self._consumer.partitions_for_topic(full_topic_name)
        if not topic_partitions:
            raise KafkaError(f"Partitions not found for topic {full_topic_name}.")
        return topic_partitions

    def seek_offset(self, topic_partition: TopicPartition, offset: int) -> None:
        """Seek the provided partition for a configured consumer group to a specific offset.

        Arguments:
            topic_partition (kafka.structs.TopicPartition): Non-localized topic partition.
            offset (int): Desired offset.

        """
        local_tp = self._get_localized_tp(topic_partition)
        if self._dry_run:
            logger.debug(
                "dry_run mode: Attempted to commit on %s:%s to offset %s.",
                local_tp.topic,
                local_tp.partition,
                offset,
            )
        else:
            self._consumer.assign([local_tp])
            self._consumer.commit({local_tp: OffsetAndMetadata(offset, None)})

    def find_offset_for_timestamp(self, topic_partition: TopicPartition, timestamp: int) -> int:
        """Find offset by approximating it with the provided timestamp.

        Arguments:
            topic_partition (kafka.structs.TopicPartition): Non-localized topic partition.
            timestamp (int): Timestamp for offset approximation.

        Returns:
            int: Approximated offset.

        """
        local_tp = self._get_localized_tp(topic_partition)
        offset_timestamp = self._consumer.offsets_for_times({local_tp: timestamp - DELTA})
        if not offset_timestamp or not offset_timestamp[local_tp]:
            raise KafkaError(
                f"Offset not found for topic {local_tp.topic}, partition {local_tp.partition}, "
                f"when seeking by timestamp {timestamp}."
            )

        return offset_timestamp[local_tp].offset

    def __enter__(self) -> "KafkaClient":
        """Returns initiated instance."""
        return self

    def __exit__(self, *_: Any) -> None:
        """Close KafkaConsumer."""
        self._consumer.close(autocommit=False)


class Kafka:
    """Kafka module, that currently allows for inter and cross cluster consumer group position transfer."""

    def __init__(self, *, kafka_config: Dict[str, Dict[str, Dict]], dry_run: bool = True):
        """Create Kafka module instance.

        Kafka config is based on a Puppet generated config.yaml in spicerack configs. At minimum, it requires a
        ssl_string defined for each participating cluster, e.g.::

          main:
            eqiad:
               brokers:
                  ssl_string: "address:port,address:port"
                  ...

        Arguments:
              kafka_config (dict): Complete, available in Puppet, kafka definition.
              dry_run (bool, optional): Enable dry run mode.

        """
        self._dry_run = dry_run
        self._kafka_config = kafka_config

    @staticmethod
    def _get_offsets(*, client: KafkaClient, topics: List[str]) -> Dict[TopicPartition, int]:
        """Retrieves offsets for given topics, mutated for given site.

        Arguments:
            client (spicerack.kafka.KafkaClient): Kafka consumer and site prefix for a given cluster.
            topics (list[str]): List of topics (without site prefixes) to get offsets for.

        Returns:
            dict[kafka.structs.TopicPartition, int]: Mapping of topic partitions to their offsets for a given consumer.

        Raises:
            spicerack.kafka.KafkaError: When local offset couldn't be located (e.g. because of no messages).

        """
        topic_partitions = {}
        for tp in Kafka._get_topic_partitions(client=client, topics=topics):
            committed_offset = client.get_committed_offset(tp)
            topic_partitions[tp] = committed_offset
        return topic_partitions

    @staticmethod
    def _get_timestamps(client: KafkaClient, topics: List[str]) -> Dict[TopicPartition, int]:
        """Retrieves timestamps for given topics, mutated for given site.

        Arguments:
            client (spicerack.kafka.KafkaClient): Kafka consumer and site prefix for a given cluster.
            topics (list[str]): List of topics (without site prefixes) to get timestamps for.

        Returns:
            dict[kafka.structs.TopicPartition, int]: Mapping of topic partitions to their timestamps for a given
            consumer.

        Raises:
            spicerack.kafka.KafkaError: When there was no message to get timestamp from.

        """
        topic_partitions = {}
        for tp in Kafka._get_topic_partitions(client=client, topics=topics):
            topic_partitions[tp] = client.get_next_timestamp(tp)
        return topic_partitions

    @staticmethod
    def _get_topic_partitions(client: KafkaClient, topics: List[str]) -> List[TopicPartition]:
        """Generates a list of topic partitions for given topic list.

        Arguments:
            client (spicerack.kafka.KafkaClient): Kafka consumer and site prefix for a given cluster.
            topics (list[str]): List of topics (without site prefixes) to get partitions for.

        Returns:
            list[kafka.structs.TopicPartition]: List of topic partitions.

        """
        topic_partitions = []
        for topic in topics:
            for p in client.partitions_for_topic(topic):
                topic_partitions.append(TopicPartition(topic, p))
        return topic_partitions

    @staticmethod
    def _set_offsets(*, client: KafkaClient, offset_data: Dict[TopicPartition, int]) -> None:
        """Sets topic partitions offsets.

        Arguments:
            client (spicerack.kafka.KafkaClient): Kafka consumer for a given cluster.
            offset_data (dict[kafka.structs.TopicPartition, int]): Mapping of topic partitions to their timestamps
            for a given consumer.

        """
        for tp, offset in offset_data.items():
            client.seek_offset(tp, offset)

    def _set_timestamps_for_topics(self, *, client: KafkaClient, timestamps: Dict[str, int]) -> None:
        """Sets topic partitions offsets, based on timestamps (minus :py:const:`spicerack.kafka.DELTA`) and topic names.

        Arguments:
            client (spicerack.kafka.KafkaClient): Kafka consumer and site prefix for a given cluster.
            timestamps (dict[str, int]): Mapping of topics to their timestamps.

        """
        tp_timestamps = {}
        for topic, timestamp in timestamps.items():
            for p in client.partitions_for_topic(topic):
                tp = TopicPartition(topic, p)
                tp_timestamps[tp] = timestamp
        self._set_timestamps(client=client, timestamps=tp_timestamps)

    @staticmethod
    def _set_timestamps(*, client: KafkaClient, timestamps: Dict[TopicPartition, int]) -> None:
        """Sets topic partitions offsets, based on timestamps (minus :py:const:`spicerack.kafka.DELTA`).

        Arguments:
            client (spicerack.kafka.KafkaClient): Kafka consumer and site prefix for a given cluster.
            timestamps (dict[kafka.structs.TopicPartition, int]): Mapping of topic partitions to their timestamps.

        Raise:
            spicerack.kafka.KafkaError: When local offset couldn't be located (e.g. because of no messages).

        """
        for tp, timestamp in timestamps.items():
            offset = client.find_offset_for_timestamp(tp, timestamp)
            client.seek_offset(tp, offset)

    def transfer_consumer_position(
        self, topics: List[str], source_consumer: ConsumerDefinition, target_consumer: ConsumerDefinition
    ) -> None:
        """Transfers position from one Kafka consumer group to another.

        Same cluster position is an offset transfer, different cluster will involve approximation
        based on the source timestamp (with :py:const:`spicerack.kafka.DELTA` ms earlier seek time).

        All topics for which the transfer will happen are assumed to use site prefixes (e.g. eqiad.mutation).

        Arguments:
            topics (list[str]): List of topics to transfer from and to, without site prefixes.
            source_consumer (spicerack.kafka.ConsumerDefinition): Consumer definition for the source consumer group.
            target_consumer (spicerack.kafka.ConsumerDefinition): Consumer definition for the target consumer group.

        """
        with KafkaClient(
            consumer_definition=source_consumer, kafka_config=self._kafka_config, dry_run=self._dry_run
        ) as source_client, KafkaClient(
            consumer_definition=target_consumer, kafka_config=self._kafka_config, dry_run=self._dry_run
        ) as target_client:

            if (source_consumer.cluster, source_consumer.site) == (target_consumer.cluster, target_consumer.site):
                logger.info("Same cluster, setting offsets...")
                offset_data = Kafka._get_offsets(client=source_client, topics=topics)
                logger.info(
                    'Extracted offsets from source cluster "%s", site "%s" and consumer group "%s".',
                    source_consumer.cluster,
                    source_consumer.site,
                    source_consumer.consumer_group,
                )
                self._set_offsets(client=target_client, offset_data=offset_data)
                logger.info(
                    'Offsets set for target cluster "%s", site "%s" and consumer group "%s".',
                    target_consumer.cluster,
                    target_consumer.site,
                    target_consumer.consumer_group,
                )

            else:
                logger.info("Different clusters, approximating offsets based on timestamps...")
                from_offsets_timestamps = Kafka._get_timestamps(client=source_client, topics=topics)
                logger.info(
                    'Extracted timestamps from source cluster "%s", site "%s" and consumer group "%s".',
                    source_consumer.cluster,
                    source_consumer.site,
                    source_consumer.consumer_group,
                )
                self._set_timestamps(client=target_client, timestamps=from_offsets_timestamps)
                logger.info(
                    'Offsets approximated and set for target cluster "%s", site "%s" and consumer group "%s".',
                    target_consumer.cluster,
                    target_consumer.site,
                    target_consumer.consumer_group,
                )

        logger.info("Done.")

    def set_consumer_position_by_timestamp(
        self, target_consumer: ConsumerDefinition, timestamps: Dict[str, int]
    ) -> None:
        """Set an approximated offsets for given topics (provided without site prefix).

        Module uses timestamps earlier by :py:const:`spicerack.kafka.DELTA` ms.

        Arguments:
            target_consumer (spicerack.kafka.ConsumerDefinition): Consumer definition for the target consumer group.
            timestamps (dict[str, int): List of topics with timestamps to use.

        """
        with KafkaClient(
            consumer_definition=target_consumer, kafka_config=self._kafka_config, dry_run=self._dry_run
        ) as client:
            logger.info("Approximating offsets based on provided timestamps...")
            self._set_timestamps_for_topics(client=client, timestamps=timestamps)

        logger.info("Done.")



"""Dnsdisc module tests."""
from datetime import timedelta
from unittest import mock

import pytest
from wmflib.exceptions import WmflibError

from spicerack.decorators import retry
from spicerack.exceptions import SpicerackError


class DryRunRetry:
    """Test class to simulate any Spicerack class with a ``_dry_run`` property."""

    def __init__(self, dry_run=True):
        """Initialize the ``_dry_run`` property."""
        self._dry_run = dry_run

    @retry
    def fail(self):
        """A method that fail always to trigger the retry logic."""
        raise SpicerackError(f"self._dry_run={self._dry_run}")


def _generate_mocked_function(calls):
    func = mock.Mock()
    func.side_effect = calls
    func.__qualname__ = "mocked"
    return func


@pytest.mark.parametrize(
    "calls, sleep_calls",
    (
        ([True], []),
        ([SpicerackError("error1"), True], [3.0]),
        ([SpicerackError("error1"), SpicerackError("error2"), True], [3.0, 9.0]),
    ),
)
@mock.patch("wmflib.decorators.time.sleep", return_value=None)
def test_retry_pass_no_args(mocked_sleep, calls, sleep_calls):
    """Using @retry with no arguments should use the default values."""
    func = _generate_mocked_function(calls)
    ret = retry(func)()
    assert ret
    func.assert_has_calls([mock.call()] * len(calls))
    mocked_sleep.assert_has_calls([mock.call(i) for i in sleep_calls])


@pytest.mark.parametrize("dry_run", (True, False))
@mock.patch("wmflib.decorators.time.sleep", return_value=None)
def test_retry_pass_no_args_dry_run(mocked_sleep, dry_run):
    """Using @retry with no arguments should use the default values but set tries to 1 if in DRY-RUN."""
    obj = DryRunRetry(dry_run=dry_run)
    with pytest.raises(SpicerackError):
        obj.fail()

    sleep_call_count = 2
    if dry_run:
        sleep_call_count = 0
    assert mocked_sleep.call_count == sleep_call_count


@mock.patch("wmflib.decorators.time.sleep", return_value=None)
def test_retry_pass_no_args_dry_run_func(mocked_sleep):
    """Using @retry with no arguments should use the default values but set tries to 1 if in DRY-RUN."""

    @retry
    def a_func(dry_run=True):
        print(dry_run)
        raise SpicerackError

    with pytest.raises(SpicerackError):
        a_func()

    assert not mocked_sleep.called


@pytest.mark.parametrize(
    "exc, calls, sleep_calls",
    (
        (SpicerackError, [SpicerackError("error")] * 3, [3.0, 9.0]),
        (Exception, [Exception("error")], []),
    ),
)
@mock.patch("wmflib.decorators.time.sleep", return_value=None)
def test_retry_fail_no_args(mocked_sleep, exc, calls, sleep_calls):
    """Using @retry with no arguments should raise the exception raised by the decorated function if not cathced."""
    func = _generate_mocked_function(calls)
    with pytest.raises(exc, match="error"):
        retry(func)()

    func.assert_has_calls([mock.call()] * len(calls))
    mocked_sleep.assert_has_calls([mock.call(i) for i in sleep_calls])


@pytest.mark.parametrize(
    "calls, sleep_calls, kwargs",
    (
        ([True], [], {"delay": timedelta(seconds=11), "tries": 1}),
        (
            [Exception("error1"), True],
            [5.55],
            {
                "delay": timedelta(seconds=5, milliseconds=550),
                "tries": 2,
                "exceptions": Exception,
            },
        ),
        (
            [SpicerackError("error1"), True],
            [8.88],
            {
                "backoff_mode": "exponential",
                "delay": timedelta(milliseconds=8880),
                "tries": 2,
            },
        ),
        (
            [SpicerackError("error1"), True],
            [0.90],
            {"backoff_mode": "power", "delay": timedelta(milliseconds=900), "tries": 2},
        ),
    ),
)
@mock.patch("wmflib.decorators.time.sleep", return_value=None)
def test_retry_pass_args(mocked_sleep, calls, sleep_calls, kwargs):
    """Using @retry with arguments should use the soecified values."""
    func = _generate_mocked_function(calls)
    ret = retry(**kwargs)(func)()

    assert ret
    func.assert_has_calls([mock.call()] * len(calls))
    mocked_sleep.assert_has_calls([mock.call(i) for i in sleep_calls])


@pytest.mark.parametrize(
    "exc, kwargs",
    (
        (SpicerackError, {}),
        (RuntimeError, {"exceptions": (KeyError, ValueError)}),
    ),
)
@mock.patch("wmflib.decorators.time.sleep", return_value=None)
def test_retry_fail_args(mocked_sleep, exc, kwargs):
    """Using @retry with arguments should raise the exception raised by the decorated function if not cathced."""
    func = _generate_mocked_function([exc("error")])
    kwargs["tries"] = 1
    with pytest.raises(exc, match="error"):
        retry(**kwargs)(func)()

    func.assert_called_once_with()
    assert not mocked_sleep.called


@pytest.mark.parametrize(
    "kwargs, message",
    (
        ({"tries": 0}, "Tries must be a positive integer, got 0"),
        ({"backoff_mode": "invalid"}, "Invalid backoff_mode: invalid"),
        (
            {"backoff_mode": "exponential", "delay": timedelta(milliseconds=500)},
            "Delay must be greater than 1 if backoff_mode is exponential",
        ),
    ),
)
def test_retry_invalid(kwargs, message):
    """Using @retry with invalid arguments should raise WmflibError."""
    with pytest.raises(WmflibError, match=message):
        retry(**kwargs)(lambda: True)()



"""Kubernetes module."""
import logging
import time
from http import HTTPStatus
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import kubernetes  # mypy: no-type
from kubernetes import client, config  # mypy: no-type
from kubernetes.client.models import V1Taint

from spicerack.decorators import retry
from spicerack.exceptions import SpicerackCheckError, SpicerackError

logger = logging.getLogger(__name__)


class KubernetesApiError(SpicerackError):
    """Custom error class for errors interacting with the kubernetes api."""


class KubernetesApiTooManyRequests(KubernetesApiError):
    """Custom error class for HTTP TooManyRequest errors when interacting with the kubernetes api."""


class KubernetesError(SpicerackError):
    """Custom error class for errors in running the kubernetes module."""


class KubernetesCheckError(SpicerackCheckError):
    """Custom error class for errors checking kubernetes resources."""


class Kubernetes:
    """High-level interface for interacting with the kubernetes api from spicerack."""

    def __init__(self, group: str, cluster: str, *, dry_run: bool = True):
        """Initialize the instance.

        Arguments:
          group (string): the cluster group we want to operate on
          cluster (string): the cluster we're operating on
          dry_run (bool): if true, no write operations will happen.

        """
        self.group = group
        self.api = KubernetesApiFactory(cluster)
        self.dry_run = dry_run

    def get_node(self, name: str) -> "KubernetesNode":
        """Get an object for a kubernetes node.

        Arguments:
          name (string): the name of the node

        Returns:
          spicerack.kubernetes.KubernetesNode

        Raises:
          spicerack.kubernetes.KubernetesApiError if the node is not found on the cluster.

        """
        return KubernetesNode(name, self.api, self.dry_run)

    def get_pod(self, namespace: str, name: str) -> "KubernetesPod":
        """Get an object for a kubernetes pod.

        Arguments:
            name (string): the name of the pod
            namespace (string): the namespace the pod is in

        Returns:
            spicerack.kubernetes.KubernetesPod

        Raises:
            spicerack.kubernetes.KubernetesApiError if the pod is not found on the cluster

        """
        return KubernetesPod(namespace, name, self.api, self.dry_run)


class KubernetesApiFactory:
    """Provides kubernetes object classes easy access to the API."""

    API_CLASSES = {"core": client.CoreV1Api}
    CONFIG_BASE = "/etc/kubernetes"

    def __init__(self, cluster: str):
        """Initialize the instace.

        Arguments:
          cluster (string): the cluster we're operating on

        """
        self.cluster = cluster
        self._configurations: Dict[str, client.Configuration] = {}

    def configuration(self, user: str) -> client.Configuration:
        """Provide the configuration for a specific user.

        Arguments:
          user (string): the user to fetch the configuration for.

        Returns:
          kubernetes.client.Configuration: the configuration for the specified user.

        Raises:
          KubernetesError if the user or the configuration are invalid

        """
        # Check the user doesn't contain path separators
        if len(Path(user).parts) != 1:
            raise KubernetesError(f"User '{user}' is not valid")
        if user not in self._configurations:
            self._configurations[user] = client.Configuration()
            try:
                config.load_kube_config(
                    config_file=str(self._config_file_path(user)), client_configuration=self._configurations[user]
                )
            except kubernetes.config.config_exception.ConfigException as e:
                raise KubernetesError(e) from e
        return self._configurations[user]

    def core(self, *, user: str = "admin") -> kubernetes.client.CoreV1Api:
        """Return an instance of the core api correctly configured.

        Arguments:
          user (string): the user to use for authentication

        """
        conf = self.configuration(user)
        return self.API_CLASSES["core"](client.ApiClient(configuration=conf))

    def _config_file_path(self, user: str) -> Path:
        """Returns the path on the configuration file.

        Returns:
          pathlib.Path the path of the configuration file

        """
        return Path(self.CONFIG_BASE) / f"{user}-{self.cluster}.config"


class KubernetesNode:
    """Encapsulates actions on a kubernetes node."""

    def __init__(
        self,
        fqdn: str,
        api: KubernetesApiFactory,
        dry_run: bool = True,
        init_obj: Optional[kubernetes.client.models.v1_node.V1Node] = None,
    ):
        """Initialize the instance.

        Arguments:
          fqdn (string): the fqdn of the node.
          api (KubernetesApiFactory): the api factory we're going to use.
          dry_run (bool): if true, no write operations will happen.
          init_obj (kubernetes.client.models.v1_node.V1Node): if not None, this api object will be used, instead of
              fetching it from the api.

        """
        self._api = api
        self._fqdn = fqdn
        self._dry_run = dry_run
        if init_obj is not None:
            if init_obj.metadata.name != self._fqdn:
                raise KubernetesError(f"Mismatched names: got {init_obj.metadata.name}, expected {self._fqdn}")
            self._node = init_obj
        else:
            # Get the object corresponding to the fqdn provided. If non-existent, it will fail.
            self._node = self._get()

    def is_schedulable(self) -> bool:
        """Checks if a node is schedulable or not.

        Returns:
          bool true if payloads can be scheduled on the node, false otherwise.

        """
        return not (self._node.spec and self._node.spec.unschedulable)

    @property
    def name(self) -> str:
        """The name of the node.

        Returns:
          string: the name of the node.

        """
        return self._node.metadata.name

    @property
    def taints(self) -> List[V1Taint]:
        """The taints of the node.

        Returns:
          List[V1Taints]: the taints of the node.

        """
        return self._node.spec.taints if self._node.spec.taints is not None else []

    def cordon(self) -> None:
        """Makes the node unschedulable.

        Raises:
          KubernetesApiError if the call to the api failed
          KubernetesCheckError if the node wasn't set to unschedulable

        """
        if not self.is_schedulable():
            logger.info("Node %s already cordoned", self.name)
            return
        if self._dry_run:
            logger.info("Would have cordoned %s", self.name)
            return
        logger.info("Cordoning %s", self.name)
        body = {"spec": {"unschedulable": True}}
        self._node = self._patch(body)
        # Now let's check if the object has changed in the api as expected.
        if self.is_schedulable():
            raise KubernetesCheckError(f"{self} is not unschedulable after trying to cordon it.")

    def uncordon(self) -> None:
        """Makes a node schedulable.

        Raises:
          KubernetesApiError if the call to the api failed
          KubernetesCheckError if the node wasn't set to unschedulable

        """
        if self.is_schedulable():
            logger.info("Node %s already schedulable", self.name)
            return
        if self._dry_run:
            logger.info("Would have uncordoned %s", self.name)
            return
        logger.info("Uncordoning %s", self.name)
        body = {"spec": {"unschedulable": False}}
        self._node = self._patch(body)
        # Now let's check if the object has changed in the api as expected.
        if not self.is_schedulable():
            raise KubernetesCheckError(f"Node {self} is not schedulable after trying to uncordon it.")

    def drain(self) -> None:
        """Drains the node, analogous to `kubectl drain`.

        Raises:
          KubernetesCheckError if we can't evict all pods

        """
        unevictable: List["KubernetesPod"] = []
        failed: List[Tuple["KubernetesPod", KubernetesApiError]] = []
        self.cordon()
        max_grace_period = 0
        for pod in self.get_pods():
            try:
                # Dry run is passed to the pods, so if we're in dry-run mode nothing will actually be evicted.
                pod.evict()
            except KubernetesError:
                # pod.evict raises a KubernetesError if the pod is unevictable
                unevictable.append(pod)
            except KubernetesApiError as e:
                failed.append((pod, e))
            # Update the max grace period.
            if not pod.is_terminated() and pod.spec.termination_grace_period_seconds > max_grace_period:
                max_grace_period = pod.spec.termination_grace_period_seconds

        if len(failed) > 0:
            for p, exc in failed:
                logger.error("Failed to evict pod %s from node %s: %s", p, self, exc)

            raise KubernetesCheckError(f"Could not evict all pods from node {self}")

        self._wait_for_empty(len(unevictable), max_grace_period)

    def refresh(self) -> None:
        """Refresh the api object from the kubernetes api server."""
        self._node = self._get()

    def _get(self) -> kubernetes.client.models.v1_node.V1Node:
        """Get a node api object.

        Arguments:
          name (string): the name of the node

        Returns:
          kubernetes.client.models.v1_node.V1Node the api object for the node.

        """
        try:
            nodes = self._api.core().list_node(field_selector=f"metadata.name={self._fqdn}")
            nodes_found = len(nodes.items)
            if nodes_found == 1:
                return nodes.items[0]
            if nodes_found == 0:  # pylint: disable=no-else-raise
                raise KubernetesError(f"Node {self._fqdn} not found")
            else:
                node_names = ",".join([o.metadata.name for o in nodes.items])
                raise KubernetesError(f"More than one node found for name {self._fqdn}: {node_names}")
        except kubernetes.client.exceptions.ApiException as exc:
            raise KubernetesApiError(f"Failed to list nodes: {exc}") from exc

    def _patch(self, body: Dict[str, Any]) -> kubernetes.client.models.v1_node.V1Node:
        """Modify the node properties."""
        try:
            return self._api.core().patch_node(self.name, body)
        except kubernetes.client.exceptions.ApiException as exc:
            raise KubernetesApiError(f"Failed to modify node: {exc}") from exc

    def get_pods(self) -> List["KubernetesPod"]:
        """Get the pods running on this node."""
        pods = []
        try:
            for obj in self._api.core().list_pod_for_all_namespaces(field_selector=f"spec.nodeName={self.name}").items:
                p = KubernetesPod(
                    obj.metadata.namespace, obj.metadata.name, self._api, dry_run=self._dry_run, init_obj=obj
                )
                pods.append(p)
            return pods
        except kubernetes.client.exceptions.ApiException as exc:
            raise KubernetesApiError(f"Failed to find pods running on node {self.name}: {exc}") from exc

    def __str__(self) -> str:
        """String representation."""
        return f"Node({self._fqdn})"

    def _wait_for_empty(self, expected: int, max_grace_period: int) -> None:
        """Wait for all pods to be evicted."""
        if self._dry_run:
            logger.info("Would have waited for node %s to be empty", self.name)
            return

        def num_pods() -> int:
            return len([p for p in self.get_pods() if not p.is_terminated()])

        @retry(
            tries=5,
            backoff_mode="exponential",
            exceptions=(KubernetesCheckError,),
            failure_message="Waiting for pods to be evicted",
        )
        def wait() -> None:
            npods = num_pods()
            if npods > expected:
                raise KubernetesCheckError(f"Node {self.name} still has {npods} pods, expected {expected}")

        # Wait for max grace period first, then retry 5 times with exponential backoff as pods need some time
        # to actually terminate and API needs some time to catch up. Especially for nodes with a large number
        # of pods.
        #
        # Please note that waiting for the max grace period is absolutely arbitrary and just what looks like
        # a reasonable time to wait for the api to conform its view of the node to reality.
        if num_pods() > expected:
            logger.debug("Waiting %d seconds before checking evictions again", max_grace_period)
            time.sleep(max_grace_period)
            wait()


class KubernetesPod:
    """Encapsulates actions on a kubernetes pod."""

    def __init__(
        self,
        namespace: str,
        name: str,
        api: KubernetesApiFactory,
        dry_run: bool = True,
        init_obj: Optional[kubernetes.client.models.v1_pod.V1Pod] = None,
    ):
        """Initialize the pod isntance.

        Arguments:
            namespace (string): the namespace where the pod is located.
            name (string): the name of the pod
            api (KubernetesApiFactory): the api factory we're going to use.
            dry_run (bool): if true, no write operations will happen.
            init_obj (kubernetes.client.models.v1_pod.V1Pod): if not None, this api object will be used, instead of
                fetching it from the api.

        """
        self._api = api
        self.name = name
        self.namespace = namespace
        self._dry_run = dry_run
        if init_obj is not None:
            if self.name != init_obj.metadata.name:
                raise KubernetesError(f"Mismatched names: got {init_obj.metadata.name}, expected {self.name}")
            if self.namespace != init_obj.metadata.namespace:
                raise KubernetesError(
                    f"Mismatched namespaces: {init_obj.metadata.namespace}, expected {self.namespace}"
                )
            self._pod = init_obj
        else:
            self._pod = self._get()

    @property
    def controller(self) -> Optional[kubernetes.client.models.v1_owner_reference.V1OwnerReference]:
        """Get the reference to the controlling object, if any.

        Returns:
          kubernetes.client.models.v1_owner_reference.V1OwnerReference: the reference.

        """
        ref = self._pod.metadata.owner_references
        if ref is None or len(ref) == 0:
            return None

        return ref[0]

    def is_daemonset(self) -> bool:
        """Checks if the pod is part of a daemonset.

        Returns:
          bool true if it is a daemonset, false otherwise.

        """
        if self.controller is None:
            return False
        return self.controller.kind == "DaemonSet"

    def is_terminated(self) -> bool:
        """Checks if the pod is terminated.

        Returns:
          bool true if the pod is not running, false otherwise.

        """
        return self._pod.status.phase in ["Succeeded", "Failed"]

    def is_mirror(self) -> bool:
        """Check if the pod is a mirror pod.

        Returns:
          bool true if this is a mirror pod, false otherwise.

        """
        return "kubernetes.io/config.mirror" in self._pod.metadata.annotations

    @property
    def spec(self) -> kubernetes.client.models.v1_pod_spec.V1PodSpec:
        """Acess the pod's spec.

        Returns:
          kubernetes.client.models.v1_pod_spec.V1PodSpec the pod spec

        """
        return self._pod.spec

    def is_evictable(self) -> bool:
        """Check if the pod can be evicted.

        Returns:
          bool  true if this pod can be evicted, false otherwise.

        """
        # We apply the logic found in kubectl:
        # https://github.com/kubernetes/kubernetes/blob/release-1.16/staging/src/k8s.io/kubectl/pkg/drain/filters.go
        # Check zero: a terminated pod is always evictable.
        if self.is_terminated():
            return True
        # Check one: the pod is orphaned and not finished
        if self.controller is None:
            logger.warning("Pod %s is orphaned, not evictable", self)
            return False
        # Check two: the pod is a daemonset
        if self.is_daemonset():
            logger.warning("Pod %s is a daemonset, not evictable", self)
            return False
        # Check three: the pod is a mirror
        if self.is_mirror():
            logger.warning("Pod %s is a mirror pod, not evictable", self)
            return False
        return True

    def evict(self) -> None:
        """Submit an eviction request to the kubernetes api for this pod.

        Raises:
          KubernetesApiTooManyRequests in case of a persistent HTTP 429 from the server.
          KubernetesApiError in case of a bad response from the server.
          KubernetesError if the pod is not evictable.

        """
        if not self.is_evictable():
            raise KubernetesError(f"Pod {self} is not evictable.")

        if self._dry_run:
            logger.info("Would have evicted %s", self)
            return
        logger.debug("Evicting pod %s", self)
        body = kubernetes.client.V1beta1Eviction(metadata=client.V1ObjectMeta(name=self.name, namespace=self.namespace))

        @retry(
            tries=5,
            backoff_mode="exponential",
            exceptions=(KubernetesApiTooManyRequests,),
            failure_message=f"Retrying eviction of {self}. API error was",
        )
        def retry_evict() -> None:
            try:
                self._api.core().create_namespaced_pod_eviction(self.name, self.namespace, body)
            except kubernetes.client.exceptions.ApiException as e:
                # The eviction is not currently allowed because of a PodDisruptionBudget or
                # we hit an API rate limit.
                # In both cases we should retry the eviction.
                if e.status == HTTPStatus.TOO_MANY_REQUESTS:
                    logger.info("Failed to evict pod %s - HTTP response body: %s", self, e.body)
                    raise KubernetesApiTooManyRequests(e) from e
                raise KubernetesApiError(e) from e

        retry_evict()

    def _get(self) -> kubernetes.client.models.v1_pod.V1Pod:
        """Get the object from the api."""
        try:
            # by convention, we have a read-only user with the same name as the namespace.
            return self._api.core(user=self.namespace).read_namespaced_pod(self.name, self.namespace)
        except kubernetes.client.exceptions.ApiException as e:
            raise KubernetesApiError(f"Error from the kubernetes api: {e}") from e

    def refresh(self) -> None:
        """Refresh the api object from the kubernetes api server."""
        self._pod = self._get()

    def __str__(self) -> str:
        """String representation."""
        return f"Pod({self.namespace}/{self.name})"



"""Cookbook module tests."""
import logging
import shutil
from pathlib import Path
from unittest import mock

import pytest

from spicerack import Spicerack, _cookbook, _menu, cookbook
from spicerack.tests import SPICERACK_TEST_PARAMS, get_fixture_path
from spicerack.tests.unit.test__log import _reset_logging_module

COOKBOOKS_BASE_PATH = Path("spicerack/tests/fixtures/cookbook")
LIST_COOKBOOKS_ALL = """cookbooks
|-- class_api
|   |-- class_api.call_another_cookbook
|   |-- class_api.example
|   |-- class_api.get_runner_raise
|   |-- class_api.multiple.CookbookA
|   |-- class_api.multiple.CookbookB
|   |-- class_api.rollback
|   |-- class_api.rollback_raise
|   |-- class_api.runtime_description
|   `-- class_api.runtime_description_raise
|-- cookbook
|-- group1
|   `-- group1.cookbook1
|-- group2
|   |-- group2.cookbook2
|   |-- group2.subgroup1
|   |   `-- group2.subgroup1.cookbook3
|   `-- group2.zcookbook4
|-- group3
|   |-- group3.argparse_ok
|   |-- group3.argument_parser_raise_system_exit
|   |-- group3.get_argument_parser_raise
|   |-- group3.get_argument_parser_raise_system_exit_str
|   |-- group3.keyboard_interrupt
|   |-- group3.non_zero_exit
|   |-- group3.raise_exception
|   |-- group3.raise_system_exit_0
|   |-- group3.raise_system_exit_9
|   |-- group3.raise_system_exit_str
|   `-- group3.subgroup3
|       `-- group3.subgroup3.cookbook4
|-- multiple.CookbookA
|-- multiple.CookbookB
`-- root
"""
LIST_COOKBOOKS_ALL_VERBOSE = """cookbooks
|-- class_api: Class API Test Cookbooks.
|   |-- class_api.call_another_cookbook: A cookbook that calls another cookbook.
|   |-- class_api.example: -
|   |-- class_api.get_runner_raise: Class API get_runner raise cookbook.
|   |-- class_api.multiple.CookbookA: Multiple cookbook classes.
|   |-- class_api.multiple.CookbookB: Multiple cookbook classes.
|   |-- class_api.rollback: Class API rollback cookbook.
|   |-- class_api.rollback_raise: Class API rollback_raise cookbook.
|   |-- class_api.runtime_description: Class API cookbook that overrides runtime_description.
|   `-- class_api.runtime_description_raise: Class API runtime_description raise cookbook.
|-- cookbook: Top level class cookbook.
|-- group1: Group1 Test Cookbooks.
|   `-- group1.cookbook1: Group1 Cookbook1.
|-- group2: -
|   |-- group2.cookbook2: Group2 Cookbook2.
|   |-- group2.subgroup1: -
|   |   `-- group2.subgroup1.cookbook3: Group2 Subgroup1 Cookbook3.
|   `-- group2.zcookbook4: UNKNOWN (unable to detect title)
|-- group3: -
|   |-- group3.argparse_ok: Group3 argparse_ok.
|   |-- group3.argument_parser_raise_system_exit: Group3 argument_parser() raise SystemExit.
|   |-- group3.get_argument_parser_raise: Group3 get argument_parser() raise.
|   |-- group3.get_argument_parser_raise_system_exit_str: Group3 get argument_parser() raise SystemExit with a string.
|   |-- group3.keyboard_interrupt: Group3 Raise KeyboardInterrupt.
|   |-- group3.non_zero_exit: Group3 Non-Zero return code.
|   |-- group3.raise_exception: Group3 Raise Exception.
|   |-- group3.raise_system_exit_0: Group3 Raise SystemExit(0).
|   |-- group3.raise_system_exit_9: Group3 Raise SystemExit(9).
|   |-- group3.raise_system_exit_str: Group3 Raise SystemExit('message').
|   `-- group3.subgroup3: -
|       `-- group3.subgroup3.cookbook4: Group3 Subgroup3 Cookbook4.
|-- multiple.CookbookA: Multiple cookbook classes.
|-- multiple.CookbookB: Multiple cookbook classes.
`-- root: Top level cookbook.
"""
LIST_COOKBOOKS_GROUP3 = """cookbooks
`-- group3
    |-- group3.argparse_ok
    |-- group3.argument_parser_raise_system_exit
    |-- group3.get_argument_parser_raise
    |-- group3.get_argument_parser_raise_system_exit_str
    |-- group3.keyboard_interrupt
    |-- group3.non_zero_exit
    |-- group3.raise_exception
    |-- group3.raise_system_exit_0
    |-- group3.raise_system_exit_9
    |-- group3.raise_system_exit_str
    `-- group3.subgroup3
        `-- group3.subgroup3.cookbook4
"""
LIST_COOKBOOKS_GROUP3_SUBGROUP3 = """cookbooks
`-- group3
    `-- group3.subgroup3
        `-- group3.subgroup3.cookbook4
"""
COOKBOOKS_MENU_TTY = """#--- cookbooks args=[] ---#
[0/9] class_api: Class API Test Cookbooks.
[NOTRUN] cookbook: Top level class cookbook.
[0/1] group1: Group1 Test Cookbooks.
[0/3] group2: -
[0/11] group3: -
[NOTRUN] multiple.CookbookA: Multiple cookbook classes.
[NOTRUN] multiple.CookbookB: Multiple cookbook classes.
[NOTRUN] root: Top level cookbook.
q - Quit
h - Help
"""
COOKBOOKS_MENU_NOTTY = """#--- cookbooks args=[] ---#
[0/9] class_api: Class API Test Cookbooks.
[NOTRUN] cookbook: Top level class cookbook.
[0/1] group1: Group1 Test Cookbooks.
[0/3] group2: -
[0/11] group3: -
[NOTRUN] multiple.CookbookA: Multiple cookbook classes.
[NOTRUN] multiple.CookbookB: Multiple cookbook classes.
[NOTRUN] root: Top level cookbook.
q - Quit
h - Help
Not a tty, exiting.
"""
COOKBOOKS_GROUP1_MENU = """#--- Group1 Test Cookbooks. args=[] ---#
[NOTRUN] cookbook1: Group1 Cookbook1.
b - Back to parent menu
h - Help
"""
COOKBOOKS_GROUP2_MENU = """#--- group2 args=[] ---#
[NOTRUN] cookbook2: Group2 Cookbook2.
[0/1] subgroup1: -
[NOTRUN] zcookbook4: UNKNOWN (unable to detect title)
b - Back to parent menu
h - Help
"""
COOKBOOKS_GROUP2_COOKBOOK2_MENU_RUN = """[{'argument': None, 'k': False}, False, False]
#--- group2 args=[] ---#
[PASS] cookbook2: Group2 Cookbook2.
[0/1] subgroup1: -
[NOTRUN] zcookbook4: UNKNOWN (unable to detect title)
b - Back to parent menu
h - Help
"""
COOKBOOKS_GROUP2_SUBGROUP1_MENU = """#--- subgroup1 args=[] ---#
[NOTRUN] cookbook3: Group2 Subgroup1 Cookbook3.
b - Back to parent menu
h - Help
"""


def test_argument_parser_converts_path():
    """It should convert a path-based cookbook into a module path."""
    argv = ["group1/cookbook1.py"]
    args = _cookbook.argument_parser().parse_args(argv)
    assert args.cookbook == "group1.cookbook1"


def test_argument_parser_keeps_module():
    """It should keep a module-based cookbook as it is."""
    argv = ["group1.cookbook1"]
    args = _cookbook.argument_parser().parse_args(argv)
    assert args.cookbook == argv[0]


def test_argument_parser_accept_empty():
    """With no args it should not fail."""
    args = _cookbook.argument_parser().parse_args([])
    assert args.cookbook == ""


def test_parse_args_list():
    """Passing -l/--list should keep the cookbook None and set list."""
    argv = ["--list"]
    args = _cookbook.argument_parser().parse_args(argv)
    assert args.list
    assert args.cookbook == ""


def test_main_wrong_instance_config(capsys):
    """If the configuration file has invalid instance_params it should print an error and exit."""
    ret = _cookbook.main(["-c", str(get_fixture_path("config_wrong_overrides.yaml")), "cookbook"])
    _, err = capsys.readouterr()
    assert ret == 1
    assert "Unable to instantiate Spicerack, check your configuration" in err
    _reset_logging_module()


def test_main_call_another_cookbook_ok(capsys):
    """It should execute the cookbook that calls another cookbook."""
    ret = _cookbook.main(
        ["-c", str(get_fixture_path("config.yaml")), "class_api.call_another_cookbook", "class_api.example"]
    )
    _, err = capsys.readouterr()
    assert ret == 0
    expected = [
        "START - Cookbook class_api.call_another_cookbook",
        "START - Cookbook class_api.example",
        "END (PASS) - Cookbook class_api.example (exit_code=0)",
        "END (PASS) - Cookbook class_api.call_another_cookbook (exit_code=0)",
    ]
    for line in expected:
        assert line in err
    _reset_logging_module()


def test_main_call_another_cookbook_not_found(capsys):
    """It should fail to call another cookbook if it doesn't exists."""
    ret = _cookbook.main(
        ["-c", str(get_fixture_path("config.yaml")), "class_api.call_another_cookbook", "class_api.not_existent"]
    )
    _, err = capsys.readouterr()
    assert ret == cookbook.EXCEPTION_RETCODE
    assert "SpicerackError: Unable to find cookbook class_api.not_existent" in err
    assert "END (FAIL) - Cookbook class_api.call_another_cookbook (exit_code=99)" in err
    _reset_logging_module()


class TestCookbookCollection:
    """Test class for the CookbookCollection class."""

    def setup_method(self):
        """Initialize the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.spicerack = Spicerack(verbose=False, dry_run=False, **SPICERACK_TEST_PARAMS)
        self.spicerack_dry_run = Spicerack(verbose=False, dry_run=True, **SPICERACK_TEST_PARAMS)
        self.spicerack_verbose = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)

    def teardown_method(self):
        """Tear down the test setting resetting the logging module."""
        _reset_logging_module()

    @pytest.mark.parametrize(
        "path_filter, verbose, expected",
        (
            ("", False, LIST_COOKBOOKS_ALL),
            ("", True, LIST_COOKBOOKS_ALL_VERBOSE),
            ("group3", False, LIST_COOKBOOKS_GROUP3),
            ("group3.subgroup3", False, LIST_COOKBOOKS_GROUP3_SUBGROUP3),
            ("group3.non_existent", False, ""),
        ),
    )
    def test_cookbooks_str(self, monkeypatch, path_filter, verbose, expected):
        """The string representation of the CookbookCollection should print all the cookbooks as a tree."""
        monkeypatch.syspath_prepend(COOKBOOKS_BASE_PATH)
        if not path_filter and not verbose:  # For the first test ensure there is no __pycache__ directory.
            shutil.rmtree(COOKBOOKS_BASE_PATH / "cookbooks" / "__pycache__", ignore_errors=True)

        spicerack = self.spicerack
        if verbose:
            spicerack = self.spicerack_verbose
        cookbooks = _cookbook.CookbookCollection(COOKBOOKS_BASE_PATH, [], spicerack, path_filter=path_filter)
        assert cookbooks.menu.get_tree() == expected

    def test_cookbooks_non_existent(self):
        """The CookbookCollection class initialized with an empty path should not collect any cookbook."""
        cookbooks = _cookbook.CookbookCollection(COOKBOOKS_BASE_PATH / "non_existent", [], self.spicerack)
        assert cookbooks.menu.get_tree() == ""

    @pytest.mark.parametrize(
        "module, err_messages, absent_err_messages, code, args",
        (
            (
                "cookbook",
                [
                    "START - Cookbook cookbook",
                    "END (PASS) - Cookbook cookbook (exit_code=0)",
                ],
                [],
                0,
                [],
            ),
            ("cookbook", [], ["START - Cookbook", "END ("], 2, ["arg1"]),
            (
                "group3.non_zero_exit",
                ["END (FAIL) - Cookbook group3.non_zero_exit (exit_code=1)"],
                [],
                1,
                [],
            ),
            (
                "group3.non_zero_exit",
                ["END (FAIL) - Cookbook group3.non_zero_exit (exit_code=1)"],
                [],
                1,
                [],
            ),
            (
                "group3.non_existent",
                ["Unable to find cookbook"],
                [],
                cookbook.NOT_FOUND_RETCODE,
                [],
            ),
            ("group3.argparse_ok", [], ["START - Cookbook", "END ("], 0, ["-h"]),
            (
                "group3.invalid_syntax",
                ["invalid syntax (invalid_syntax.py, line 7)"],
                [],
                cookbook.NOT_FOUND_RETCODE,
                [],
            ),
            (
                "group3.keyboard_interrupt",
                ["Ctrl+c pressed"],
                [],
                cookbook.INTERRUPTED_RETCODE,
                [],
            ),
            (
                "group3.get_argument_parser_raise",
                ["raised while getting argument parser for cookbook"],
                ["START - Cookbook", "END ("],
                cookbook.GET_ARGS_PARSER_FAIL_RETCODE,
                [],
            ),
            (
                "group3.get_argument_parser_raise_system_exit_str",
                ["SystemExit('argument_parser')"],
                ["START - Cookbook", "END ("],
                cookbook.GET_ARGS_PARSER_FAIL_RETCODE,
                [],
            ),
            (
                "group3.argument_parser_raise_system_exit",
                [],
                ["group3.argument_parser_raise_system_exit", "START - Cookbook", "END ("],
                2,
                [],
            ),
            (
                "group3.raise_exception",
                ["Exception: Something went wrong"],
                [],
                cookbook.EXCEPTION_RETCODE,
                [],
            ),
            ("group3.raise_system_exit_0", ["SystemExit(0) raised"], [], 0, []),
            ("group3.raise_system_exit_9", ["SystemExit(9) raised"], [], 9, []),
            (
                "group3.raise_system_exit_str",
                ["SystemExit('message') raised"],
                [],
                cookbook.EXCEPTION_RETCODE,
                [],
            ),
            (
                "class_api.example",
                [
                    "START - Cookbook class_api.example",
                    "END (PASS) - Cookbook class_api.example (exit_code=0)",
                ],
                [],
                0,
                [],
            ),
            (
                "class_api.multiple.CookbookA",
                [
                    "START - Cookbook class_api.multiple.CookbookA",
                    "END (PASS) - Cookbook class_api.multiple.CookbookA (exit_code=0)",
                ],
                [],
                0,
                [],
            ),
            (
                "class_api.get_runner_raise",
                [],
                ["START - Cookbook", "END ("],
                cookbook.CLASS_FAIL_INIT_RETCODE,
                [],
            ),
            (
                "class_api.runtime_description",
                [
                    "START - Cookbook class_api.runtime_description Runtime description",
                    "END (PASS) - Cookbook class_api.runtime_description (exit_code=0) Runtime description",
                ],
                [],
                0,
                [],
            ),
            (
                "class_api.runtime_description_raise",
                [
                    "START - Cookbook class_api.runtime_description_raise",
                    "END (PASS) - Cookbook class_api.runtime_description_raise (exit_code=0)",
                ],
                [],
                0,
                [],
            ),
            (
                "class_api.rollback",
                [
                    "START - Cookbook class_api.rollback",
                    "run has raised",
                    "rollback called",
                    "END (FAIL) - Cookbook class_api.rollback (exit_code=99)",
                ],
                [],
                cookbook.EXCEPTION_RETCODE,
                [],
            ),
            (
                "class_api.rollback_raise",
                [
                    "START - Cookbook class_api.rollback_raise",
                    "rollback has raised",
                    "END (FAIL) - Cookbook class_api.rollback_raise (exit_code=93)",
                ],
                [],
                cookbook.ROLLBACK_FAIL_RETCODE,
                [],
            ),
        ),
    )
    def test_main_execute_cookbook(  # pylint: disable=too-many-arguments
        self, tmpdir, caplog, module, err_messages, absent_err_messages, code, args
    ):
        """Calling main with the given cookbook and args should execute it."""
        config = {
            "cookbooks_base_dir": COOKBOOKS_BASE_PATH,
            "logs_base_dir": tmpdir.strpath,
        }
        with mock.patch("spicerack._cookbook.load_yaml_config", lambda config_dir: config):
            with mock.patch("spicerack._cookbook.Spicerack", return_value=self.spicerack):
                with caplog.at_level(logging.INFO):
                    ret = _cookbook.main([module] + args)

        assert ret == code
        for message in err_messages:
            assert message in caplog.text
        for message in absent_err_messages:
            assert message not in caplog.text

    def test_main_execute_cookbook_invalid_args(self, tmpdir, capsys, caplog):
        """Calling a cookbook with the wrong args should let argparse print its message."""
        config = {
            "cookbooks_base_dir": COOKBOOKS_BASE_PATH,
            "logs_base_dir": tmpdir.strpath,
        }
        with mock.patch("spicerack._cookbook.load_yaml_config", lambda config_dir: config):
            with mock.patch("spicerack._cookbook.Spicerack", return_value=self.spicerack):
                with caplog.at_level(logging.INFO):
                    ret = _cookbook.main(["group3.argparse_ok", "--invalid"])

        assert ret == 2
        _, err = capsys.readouterr()
        assert "group3.argparse_ok: error: unrecognized arguments" in err
        for message in ("START - Cookbook", "END ("):
            assert message not in caplog.text

    def test_main_execute_dry_run(self, capsys, tmpdir):
        """Calling main() with a cookbook and dry_run mode should execute it and set the dry run mode."""
        config = {
            "cookbooks_base_dir": COOKBOOKS_BASE_PATH,
            "logs_base_dir": tmpdir.strpath,
        }
        with mock.patch("spicerack._cookbook.load_yaml_config", lambda config_dir: config):
            with mock.patch("spicerack._cookbook.Spicerack", return_value=self.spicerack_dry_run):
                ret = _cookbook.main(["-d", "root"])

        assert ret == 0
        _, err = capsys.readouterr()
        assert "DRY-RUN" in err

    def test_main_list(self, tmpdir, capsys, caplog):
        """Calling main() with the -l/--list option should print the available cookbooks."""
        config = {
            "cookbooks_base_dir": COOKBOOKS_BASE_PATH,
            "logs_base_dir": tmpdir.strpath,
        }
        with mock.patch("spicerack._cookbook.load_yaml_config", lambda config_dir: config):
            with mock.patch("spicerack._cookbook.Spicerack", return_value=self.spicerack):
                with caplog.at_level(logging.INFO):
                    ret = _cookbook.main(["-l"])

        out, _ = capsys.readouterr()

        assert ret == 0
        assert out == LIST_COOKBOOKS_ALL
        lines = [
            "Failed to import module cookbooks.group3.invalid_syntax: invalid syntax (invalid_syntax.py, line 7)",
            "Failed to import module cookbooks.group3.invalid_subgroup: invalid syntax (__init__.py, line 2)",
        ]
        for line in lines:
            assert line in caplog.text

    def test_cookbooks_menu_status(self, monkeypatch):
        """Calling status on a TreeItem should show the completed and total tasks."""
        monkeypatch.syspath_prepend(COOKBOOKS_BASE_PATH)
        cookbooks = _cookbook.CookbookCollection(COOKBOOKS_BASE_PATH, [], self.spicerack)
        menu = cookbooks.get_item("")
        assert menu.status == "0/28"

    def test_cookbooks_menu_status_done(self, monkeypatch):
        """Calling status on a TreeItem with all tasks completed should return DONE."""
        monkeypatch.syspath_prepend(COOKBOOKS_BASE_PATH)
        cookbooks = _cookbook.CookbookCollection(COOKBOOKS_BASE_PATH, [], self.spicerack, path_filter="group1")
        menu = cookbooks.get_item("group1")
        assert menu.status == "0/1"
        item = cookbooks.get_item("group1.cookbook1")
        item.run()
        assert menu.status == "DONE"

    @mock.patch("spicerack._cookbook.CookbookItem", side_effect=_menu.MenuError("fail to init"))
    def test_cookbooks_menu_cookbook_init_fail(self, mocked_cookbook_item, caplog, monkeypatch):
        """When a CookbookItem object fail to get initialized it should catch the MenuError, log it and continue."""
        monkeypatch.syspath_prepend(COOKBOOKS_BASE_PATH)
        with caplog.at_level(logging.INFO):
            cookbooks = _cookbook.CookbookCollection(COOKBOOKS_BASE_PATH, [], self.spicerack, path_filter="cookbook")
            menu = cookbooks.get_item(".".join((cookbooks.cookbooks_module_prefix, "group1")))
        assert menu is None
        assert "fail to init" in caplog.text
        assert mocked_cookbook_item.called

    def test_cookbooks_cookbook_init_fail(self):
        """When a CookbookItem object fail to get initialized it should catch the MenuError, log it and continue."""

        class NotCookbookBaseSubclass:
            """Just a class that doesn't inherit from CookbookBase."""

        with pytest.raises(_menu.MenuError, match="is not a subclass of CookbookBase"):
            _menu.CookbookItem(NotCookbookBaseSubclass, [], self.spicerack)

    @pytest.mark.parametrize(
        "tty, answer, output",
        (
            (False, "q", COOKBOOKS_MENU_NOTTY),
            (True, "q", COOKBOOKS_MENU_TTY),
            (True, KeyboardInterrupt, COOKBOOKS_MENU_TTY + "QUIT\n"),
            (True, ["", "q"], COOKBOOKS_MENU_TTY + COOKBOOKS_MENU_TTY),
            (
                True,
                ["invalid", "q"],
                COOKBOOKS_MENU_TTY + "==> Invalid input <==\n" + COOKBOOKS_MENU_TTY,
            ),
            (
                True,
                ["group1", "b", "q"],
                COOKBOOKS_MENU_TTY + COOKBOOKS_GROUP1_MENU + COOKBOOKS_MENU_TTY,
            ),
            (
                True,
                ["group2", "cookbook2", "b", "q"],
                COOKBOOKS_MENU_TTY
                + COOKBOOKS_GROUP2_MENU
                + COOKBOOKS_GROUP2_COOKBOOK2_MENU_RUN
                + COOKBOOKS_MENU_TTY.replace("[0/3] group2", "[1/3] group2"),
            ),
            (
                True,
                ["group2", "cookbook2 --argument value", "b", "q"],
                COOKBOOKS_MENU_TTY
                + COOKBOOKS_GROUP2_MENU
                + COOKBOOKS_GROUP2_COOKBOOK2_MENU_RUN.replace("'argument': None", "'argument': 'value'")
                + COOKBOOKS_MENU_TTY.replace("[0/3] group2", "[1/3] group2"),
            ),
            (
                True,
                ["group2 -k", "cookbook2", "b", "q"],
                COOKBOOKS_MENU_TTY
                + COOKBOOKS_GROUP2_MENU.replace("[]", "['-k']")
                + COOKBOOKS_GROUP2_COOKBOOK2_MENU_RUN.replace("[]", "['-k']").replace("'k': False", "'k': True")
                + COOKBOOKS_MENU_TTY.replace("[0/3] group2", "[1/3] group2"),
            ),
            (
                True,
                "h",
                COOKBOOKS_MENU_TTY
                + _menu.HELP_MESSAGE.format(statuses=_cookbook.CookbookItem.statuses)
                + "\n"
                + COOKBOOKS_MENU_TTY,
            ),
        ),
    )
    @mock.patch("spicerack._cookbook.sys.stdout.isatty")
    @mock.patch("builtins.input")
    def test_cookbooks_main_menu(
        self, mocked_input, mocked_tty, tty, answer, output, capsys, tmpdir
    ):  # pylint: disable=too-many-arguments
        """Calling main() with a menu should show the menu and allow to interact with it."""
        mocked_tty.return_value = tty
        mocked_input.side_effect = answer
        config = {
            "cookbooks_base_dir": COOKBOOKS_BASE_PATH,
            "logs_base_dir": tmpdir.strpath,
        }
        with mock.patch("spicerack._cookbook.load_yaml_config", lambda config_dir: config):
            with mock.patch("spicerack._cookbook.Spicerack", return_value=self.spicerack):
                ret = _cookbook.main([])

        out, _ = capsys.readouterr()
        assert ret == 0
        assert out == output

    @mock.patch("spicerack._cookbook.sys.stdout.isatty")
    @mock.patch("builtins.input")
    def test_cookbooks_submenu(self, mocked_input, mocked_tty, capsys, tmpdir):
        """Calling main() with a menu should show the menu and allow to interact with it."""
        mocked_tty.return_value = True
        mocked_input.side_effect = ["subgroup1", "b", "q"]
        config = {
            "cookbooks_base_dir": COOKBOOKS_BASE_PATH,
            "logs_base_dir": tmpdir.strpath,
        }
        with mock.patch("spicerack._cookbook.load_yaml_config", lambda config_dir: config):
            with mock.patch("spicerack._cookbook.Spicerack", return_value=self.spicerack):
                ret = _cookbook.main(["group2"])

        out, _ = capsys.readouterr()
        assert ret == 0
        group2 = COOKBOOKS_GROUP2_MENU.replace("b - Back to parent menu", "q - Quit")
        assert out == group2 + COOKBOOKS_GROUP2_SUBGROUP1_MENU + group2
